"""This file declares the functions and datatypes of the sick_scan_xd python-API.

See doc/sick_scan_api/sick_scan_api.md for further information.

"""

import ctypes
import numpy as np
import os
from enum import Enum

""" 
Message definitions
""" 

class SickScanHeader(ctypes.Structure):
    """ 
    sick_scan_api: struct SickScanHeader, equivalent to ros::std_msgs::Header

    Attributes
    ----------
    seq : ctypes.c_uint32
        sequence ID: consecutively increasing ID
    timestamp_sec : ctypes.c_uint32
        seconds part of message timestamps: seconds (stamp_secs) since epoch (in Python the variable is called 'secs')
    timestamp_nsec : ctypes.c_uint32
        seconds part of message timestamps: nanoseconds since stamp_secs (in Python the variable is called 'nsecs')
    frame_id : ctypes.c_char * 256
        Frame this data is associated with
    """
    _fields_ = [
        ("seq", ctypes.c_uint32),                    # sequence ID: consecutively increasing ID
        ("timestamp_sec", ctypes.c_uint32),          # seconds part of message timestamps: seconds (stamp_secs) since epoch (in Python the variable is called 'secs')
        ("timestamp_nsec", ctypes.c_uint32),         # seconds part of message timestamps: nanoseconds since stamp_secs (in Python the variable is called 'nsecs')
        ("frame_id", ctypes.c_char * 256)            # Frame this data is associated with
    ]

class SickScanUint8Array(ctypes.Structure):
    """ 
    equivalent to sick_scan_api.h struct SickScanUint8Array

    Attributes
    ----------
    capacity : ctypes.c_uint64
        Number of allocated elements, i.e. max. number of elements in buffer, allocated buffer size is capacity * sizeof(SickScanPointFieldMsg)
    size : ctypes.c_uint64
        Number of currently used elements in the buffer
    buffer : ctypes.POINTER(ctypes.c_uint8)
        Memory, data in plain order and system endianess (buffer == 0, if size == 0 && capacity == 0, otherwise allocated memory), allocation/deallocation always managed by the caller.
    """
    _fields_ = [
        ("capacity", ctypes.c_uint64),               # Number of allocated elements, i.e. max. number of elements in buffer, allocated buffer size is capacity * sizeof(SickScanPointFieldMsg)
        ("size", ctypes.c_uint64),                   # Number of currently used elements in the buffer
        ("buffer", ctypes.POINTER(ctypes.c_uint8))   # Memory, data in plain order and system endianess (buffer == 0, if size == 0 && capacity == 0, otherwise allocated memory), allocation/deallocation always managed by the caller.
    ]

class SickScanNativeDataType(Enum):
    """ 
    This message holds the description of one point entry in the PointCloud2 message format, equivalent to type enum im ros::sensor_msgs::PointField
    """
    SICK_SCAN_POINTFIELD_DATATYPE_INT8    = 1, "SICK_SCAN_POINTFIELD_DATATYPE_INT8"
    SICK_SCAN_POINTFIELD_DATATYPE_UINT8   = 2, "SICK_SCAN_POINTFIELD_DATATYPE_UINT8"
    SICK_SCAN_POINTFIELD_DATATYPE_INT16   = 3, "SICK_SCAN_POINTFIELD_DATATYPE_INT16"
    SICK_SCAN_POINTFIELD_DATATYPE_UINT16  = 4, "SICK_SCAN_POINTFIELD_DATATYPE_UINT16"
    SICK_SCAN_POINTFIELD_DATATYPE_INT32   = 5, "SICK_SCAN_POINTFIELD_DATATYPE_INT32"
    SICK_SCAN_POINTFIELD_DATATYPE_UINT32  = 6, "SICK_SCAN_POINTFIELD_DATATYPE_UINT32"
    SICK_SCAN_POINTFIELD_DATATYPE_FLOAT32 = 7, "SICK_SCAN_POINTFIELD_DATATYPE_FLOAT32"
    SICK_SCAN_POINTFIELD_DATATYPE_FLOAT64 = 8, "SICK_SCAN_POINTFIELD_DATATYPE_FLOAT64"
    def __int__(self):
        return self.value[0]
    def __str__(self):
        return self.value[1]


class SickScanPointFieldMsg(ctypes.Structure):
    """ 
    sick_scan_api: struct SickScanPointFieldMsg, equivalent to ros::sensor_msgs::PointField
    SickScanPointFieldArray is an array of SickScanPointFieldMsg, which defines the structure of the binary data of a SickScanPointCloudMsg.
    SickScanPointFieldMsg for pointclouds in cartesian coordinates with fields (x, y, z, intensity):
        [ SickScanPointFieldMsg(name="x", offset=0, datatype=FLOAT32, count=1), 
          SickScanPointFieldMsg(name="y", offset=4, datatype=FLOAT32, count=1),
          SickScanPointFieldMsg(name="z", offset=8, datatype=FLOAT32, count=1),
          SickScanPointFieldMsg(name="intensity", offset=12, datatype=FLOAT32, count=1) ]
    SickScanPointFieldMsg for pointclouds in polar coordinates with fields (range, azimuth, elevation, intensity):
        [ SickScanPointFieldMsg(name="range", offset=0, datatype=FLOAT32, count=1), 
          SickScanPointFieldMsg(name="azimuth", offset=4, datatype=FLOAT32, count=1),
          SickScanPointFieldMsg(name="elevation", offset=8, datatype=FLOAT32, count=1),
          SickScanPointFieldMsg(name="intensity", offset=12, datatype=FLOAT32, count=1) ]

    Attributes
    ----------
    name : ctypes.c_char * 256
        Name of field (max. length 256 characters)
    offset : ctypes.c_uint32
        Offset from start of point struct
    datatype : ctypes.c_uint8
        Datatype enumeration, see SickScanNativeDataType, equivalent to type enum im ros::sensor_msgs::PointField
    count : ctypes.c_uint32
        How many elements in the field
    """
    _fields_ = [
        ("name", ctypes.c_char * 256),               # Name of field (max. length 256 characters)
        ("offset", ctypes.c_uint32),                 # Offset from start of point struct
        ("datatype", ctypes.c_uint8),                # Datatype enumeration, see SickScanNativeDataType, equivalent to type enum im ros::sensor_msgs::PointField
        ("count", ctypes.c_uint32)                   # How many elements in the field
   ]

class SickScanPointFieldArray(ctypes.Structure):
    """ 
    sick_scan_api: struct SickScanPointFieldArray, Array of SickScanPointFieldMsg, which can be serialized and imported in C, C++ or python

    Attributes
    ----------
    capacity : ctypes.c_uint64
        Number of allocated elements, i.e. max. number of elements in buffer, allocated buffer size is capacity * sizeof(SickScanPointFieldMsg)
    size : ctypes.c_uint64
        Number of currently used elements in the buffer
    buffer : ctypes.POINTER(SickScanPointFieldMsg)
        Memory, data in plain order and system endianess (buffer == 0, if size == 0 && capacity == 0, otherwise allocated memory), allocation/deallocation always managed by the caller.
    """
    _fields_ = [
        ("capacity", ctypes.c_uint64),                     # Number of allocated elements, i.e. max. number of elements in buffer, allocated buffer size is capacity * sizeof(SickScanPointFieldMsg)
        ("size", ctypes.c_uint64),                         # Number of currently used elements in the buffer
        ("buffer", ctypes.POINTER(SickScanPointFieldMsg))  # Memory, data in plain order and system endianess (buffer == 0, if size == 0 && capacity == 0, otherwise allocated memory), allocation/deallocation always managed by the caller.
    ]

class SickScanPointCloudMsg(ctypes.Structure):
    """ 
    sick_scan_api: struct SickScanPointCloudMsg, equivalent to ros::std_msgs::PointCloud2
    SickScanPointCloudMsg contains the polar or cartesian pointcloud data.
    A SickScanPointCloudMsg in cartesian coordinates has fields (x, y, z, intensity).
    A SickScanPointCloudMsg in polar coordinates has fields (range, azimuth, elevation, intensity).
    Note: The pointcloud contains always <num_echos> echos. Invalid echos are filled with 0 values in the data array.

    Attributes
    ----------
    header : SickScanHeader
        message timestamp
    height : ctypes.c_uint32
        2D structure of the point cloud. If the cloud is unordered, height is 1
    width : ctypes.c_uint32
        and width is the length of the point cloud.
    fields : SickScanPointFieldArray
        Describes the channels and their layout in the binary data blob.
    is_bigendian : ctypes.c_uint8
        Is this data bigendian?
    point_step : ctypes.c_uint32
        Length of a point in bytes
    row_step : ctypes.c_uint32
        Length of a row in bytes
    data : SickScanUint8Array
        Actual point data, size is (row_step*height)
    is_dense : ctypes.c_uint8
        True if there are no invalid points
    num_echos : ctypes.c_int32
        number of echos
    segment_idx : ctypes.c_int32
        segment index (or -1 if pointcloud contains data from multiple segments)
    """
    _fields_ = [
        ("header", SickScanHeader),                  # message timestamp
        ("height", ctypes.c_uint32),                 # 2D structure of the point cloud. If the cloud is unordered, height is 1
        ("width", ctypes.c_uint32),                  # and width is the length of the point cloud.
        ("fields", SickScanPointFieldArray),         # Describes the channels and their layout in the binary data blob.
        ("is_bigendian", ctypes.c_uint8),            # Is this data bigendian?
        ("point_step", ctypes.c_uint32),             # Length of a point in bytes
        ("row_step", ctypes.c_uint32),               # Length of a row in bytes
        ("data", SickScanUint8Array),                # Actual point data, size is (row_step*height)
        ("is_dense", ctypes.c_uint8),                # True if there are no invalid points
        ("num_echos", ctypes.c_int32),               # number of echos
        ("segment_idx", ctypes.c_int32),             # segment index (or -1 if pointcloud contains data from multiple segments)
        ("topic", ctypes.c_char * 256)               # ros topic this pointcloud is published
   ]

class SickScanVector3Msg(ctypes.Structure):
    """ 
    sick_scan_api: struct SickScanVector3Msg, equivalent to geometry_msgs/Vector3

    Attributes
    ----------
    x : ctypes.c_double
    y : ctypes.c_double
    z : ctypes.c_double
    """
    _fields_ = [
        ("x", ctypes.c_double),
        ("y", ctypes.c_double),
        ("z", ctypes.c_double)
    ]

class SickScanQuaternionMsg(ctypes.Structure):
    """ 
    sick_scan_api: struct SickScanQuaternionMsg, equivalent to geometry_msgs/Quaternion

    Attributes
    ----------
    x : ctypes.c_double
    y : ctypes.c_double
    z : ctypes.c_double
    w : ctypes.c_double
    """
    _fields_ = [
        ("x", ctypes.c_double),
        ("y", ctypes.c_double),
        ("z", ctypes.c_double),
        ("w", ctypes.c_double)
    ]

class SickScanPointArray(ctypes.Structure):
    """ 
    sick_scan_api: struct SickScanPointArray, Array of SickScanVector3Msg, which can be serialized and imported in C, C++ or python

    Attributes
    ----------
    capacity : ctypes.c_uint64
        Number of allocated elements, i.e. max. number of elements in buffer, allocated buffer size is capacity * sizeof(SickScanVector3Msg)
    size : ctypes.c_uint64
        Number of currently used elements in the buffer
    buffer : ctypes.POINTER(SickScanVector3Msg)
        Memory, data in plain order and system endianess (buffer == 0, if size == 0 && capacity == 0, otherwise allocated memory), allocation/deallocation always managed by the caller.
    """
    _fields_ = [
        ("capacity", ctypes.c_uint64), # Number of allocated elements, i.e. max. number of elements in buffer, allocated buffer size is capacity * sizeof(SickScanVector3Msg)
        ("size", ctypes.c_uint64),     # Number of currently used elements in the buffer
        ("buffer", ctypes.POINTER(SickScanVector3Msg))  # Memory, data in plain order and system endianess (buffer == 0, if size == 0 && capacity == 0, otherwise allocated memory), allocation/deallocation always managed by the caller.
    ]

class SickScanImuMsg(ctypes.Structure):
    """ 
    sick_scan_api: struct SickScanImuMsg, equivalent to ros sensor_msgs::Imu

    Attributes
    ----------
    header : SickScanHeader
        message timestamp
    orientation : SickScanQuaternionMsg
    orientation_covariance : ctypes.c_double * 9
        Row major about x, y, z axes
    angular_velocity : SickScanVector3Msg
    angular_velocity_covariance : ctypes.c_double * 9
        Row major about x, y, z axes
    linear_acceleration : SickScanVector3Msg
    linear_acceleration_covariance : ctypes.c_double * 9
        Row major x, y z
    """
    _fields_ = [
        ("header", SickScanHeader),                              # message timestamp
        ("orientation", SickScanQuaternionMsg),
        ("orientation_covariance", ctypes.c_double * 9),         # Row major about x, y, z axes
        ("angular_velocity", SickScanVector3Msg),
        ("angular_velocity_covariance", ctypes.c_double * 9),    # Row major about x, y, z axes
        ("linear_acceleration", SickScanVector3Msg),
        ("linear_acceleration_covariance", ctypes.c_double * 9)  # Row major x, y z
    ]

class SickScanLFErecFieldMsg(ctypes.Structure):
    """ 
    sick_scan_api: struct SickScanLFErecFieldMsg, equivalent to LFErecFieldMsg.msg

    Attributes
    ----------
    version_number : ctypes.c_uint16
    field_index : ctypes.c_uint8
    sys_count : ctypes.c_uint32
    dist_scale_factor : ctypes.c_float
    dist_scale_offset : ctypes.c_float
    angle_scale_factor : ctypes.c_uint32
    angle_scale_offset : ctypes.c_int32
    field_result_mrs : ctypes.c_uint8
        field result is 0 (invalid/incorrect), 1 (free/clear) or 2 (infringed)
    time_state : ctypes.c_uint16
        No time data: 0, Time data: 1
    year : ctypes.c_uint16
        f.e. 2021
    month : ctypes.c_uint8
        1 ... 12
    day : ctypes.c_uint8
        1 ... 31
    hour : ctypes.c_uint8
        0 ... 23
    minute : ctypes.c_uint8
        0 ... 59
    second : ctypes.c_uint8
        0 ... 59
    microsecond : ctypes.c_uint32
        0 ... 999999
    """
    _fields_ = [
        ("version_number", ctypes.c_uint16),
        ("field_index", ctypes.c_uint8),
        ("sys_count", ctypes.c_uint32),
        ("dist_scale_factor", ctypes.c_float),
        ("dist_scale_offset", ctypes.c_float),
        ("angle_scale_factor", ctypes.c_uint32),
        ("angle_scale_offset", ctypes.c_int32),
        ("field_result_mrs", ctypes.c_uint8),        # field result is 0 (invalid/incorrect), 1 (free/clear) or 2 (infringed)
        ("time_state", ctypes.c_uint16),             # No time data: 0, Time data: 1
        ("year", ctypes.c_uint16),                   # f.e. 2021
        ("month", ctypes.c_uint8),                   # 1 ... 12
        ("day", ctypes.c_uint8),                     # 1 ... 31
        ("hour", ctypes.c_uint8),                    # 0 ... 23
        ("minute", ctypes.c_uint8),                  # 0 ... 59
        ("second", ctypes.c_uint8),                  # 0 ... 59
        ("microsecond", ctypes.c_uint32)             # 0 ... 999999
    ]

class SickScanLFErecMsg(ctypes.Structure):
    """ 
    sick_scan_api: struct SickScanLFErecMsg, equivalent to LFErecMsg.msg

    Attributes
    ----------
    header : SickScanHeader
        message timestamp
    fields_number : ctypes.c_uint16
        number of valid fields
    fields : SickScanLFErecFieldMsg * 3
        max. 3 valid fields
    """
    _fields_ = [
        ("header", SickScanHeader),                  # message timestamp
        ("fields_number", ctypes.c_uint16),          # number of valid fields
        ("fields", SickScanLFErecFieldMsg * 3)       # max. 3 valid fields
    ]

class SickScanLIDoutputstateMsg(ctypes.Structure):
    """ 
    sick_scan_api: struct SickScanLIDoutputstateMsg, equivalent to LIDoutputstateMsg.msg

    Attributes
    ----------
    header : SickScanHeader
        message timestamp
    version_number : ctypes.c_uint16
        Status code version number
    system_counter : ctypes.c_uint32
        Status code system counter (time in microsec since power up, max. 71 min then starting from 0 again)
    output_state : ctypes.c_uint8 * 8
        array of max. 8 output states, each state with value 0 (not active), 1 (active) or 2 (not used)
    output_count : ctypes.c_uint32 * 8
        array of max. 8 output counter
    time_state : ctypes.c_uint16
        No time data: 0, Time data: 1 (sensortime from the last change of min. one of the outputs)
    year : ctypes.c_uint16
        f.e. 2021
    month : ctypes.c_uint8
        1 ... 12
    day : ctypes.c_uint8
        1 ... 31
    hour : ctypes.c_uint8
        0 ... 23
    minute : ctypes.c_uint8
        0 ... 59
    second : ctypes.c_uint8
        0 ... 59
    microsecond : ctypes.c_uint32
        0 ... 999999
    """
    _fields_ = [
        ("header", SickScanHeader),                  # message timestamp
        ("version_number", ctypes.c_uint16),         # Status code version number
        ("system_counter", ctypes.c_uint32),         # Status code system counter (time in microsec since power up, max. 71 min then starting from 0 again)
        ("output_state", ctypes.c_uint8 * 8),        # array of max. 8 output states, each state with value 0 (not active), 1 (active) or 2 (not used)
        ("output_count", ctypes.c_uint32 * 8),       # array of max. 8 output counter
        ("time_state", ctypes.c_uint16),             # No time data: 0, Time data: 1 (sensortime from the last change of min. one of the outputs)
        ("year", ctypes.c_uint16),                   # f.e. 2021
        ("month", ctypes.c_uint8),                   # 1 ... 12
        ("day", ctypes.c_uint8),                     # 1 ... 31
        ("hour", ctypes.c_uint8),                    # 0 ... 23
        ("minute", ctypes.c_uint8),                  # 0 ... 59
        ("second", ctypes.c_uint8),                  # 0 ... 59
        ("microsecond", ctypes.c_uint32)             # 0 ... 999999
    ]

class SickScanRadarPreHeader(ctypes.Structure):
    """ 
    sick_scan_api: struct SickScanRadarPreHeader, equivalent to RadarPreHeader.msg

    Attributes
    ----------
    uiversionno : ctypes.c_uint16
        version number
    sick_scan/RadarPreHeaderDeviceBlock:
    uiident : ctypes.c_uint32
        Logical number of the device"
    udiserialno : ctypes.c_uint32
        Serial number of the device
    bdeviceerror : ctypes.c_uint8
        State of the device
    bcontaminationwarning : ctypes.c_uint8
        Contamination Warning
    bcontaminationerror : ctypes.c_uint8
        Contamination Error
    sick_scan/RadarPreHeaderStatusBlock:
    uitelegramcount : ctypes.c_uint32
        telegram number
    uicyclecount : ctypes.c_uint32
        "scan number"
    udisystemcountscan : ctypes.c_uint32
        system time since power on of scan gen. [us]
    udisystemcounttransmit : ctypes.c_uint32
        system time since power on of scan transmission [us]
    uiinputs : ctypes.c_uint16
        state of digital inputs
    uioutputs : ctypes.c_uint16
        state of digital outputs
    sick_scan/RadarPreHeaderMeasurementParam1Block:
    uicycleduration : ctypes.c_uint32
        Time in microseconds to detect this items
    uinoiselevel : ctypes.c_uint32
        [1/100dB]
    sick_scan/RadarPreHeaderEncoderBlock[]:
    numencoder : ctypes.c_uint16
        number of valid encoders (0)
    udiencoderpos : ctypes.c_uint32 * 3
        array of max. 3 encoder position [inc]
    iencoderspeed : ctypes.c_uint16 * 3
        array of max. 3 encoder speed [inc/mm]
    """
    _fields_ = [
        ("uiversionno", ctypes.c_uint16),            # version number
        # sick_scan/RadarPreHeaderDeviceBlock:
        ("uiident", ctypes.c_uint32),                # Logical number of the device"
        ("udiserialno", ctypes.c_uint32),            # Serial number of the device
        ("bdeviceerror", ctypes.c_uint8),            # State of the device
        ("bcontaminationwarning", ctypes.c_uint8),   # Contamination Warning
        ("bcontaminationerror", ctypes.c_uint8),     # Contamination Error
        # sick_scan/RadarPreHeaderStatusBlock:
        ("uitelegramcount", ctypes.c_uint32),        # telegram number
        ("uicyclecount", ctypes.c_uint32),           # "scan number"
        ("udisystemcountscan", ctypes.c_uint32),     # system time since power on of scan gen. [us]
        ("udisystemcounttransmit", ctypes.c_uint32), # system time since power on of scan transmission [us]
        ("uiinputs", ctypes.c_uint16),               # state of digital inputs
        ("uioutputs", ctypes.c_uint16),              # state of digital outputs
        # sick_scan/RadarPreHeaderMeasurementParam1Block:
        ("uicycleduration", ctypes.c_uint32),        # Time in microseconds to detect this items
        ("uinoiselevel", ctypes.c_uint32),           # [1/100dB]
        # sick_scan/RadarPreHeaderEncoderBlock[]:
        ("numencoder", ctypes.c_uint16),             # number of valid encoders (0)
        ("udiencoderpos", ctypes.c_uint32 * 3),      # array of max. 3 encoder position [inc]
        ("iencoderspeed", ctypes.c_uint16 * 3)       # array of max. 3 encoder speed [inc/mm]
    ]

class SickScanRadarObject(ctypes.Structure):
    """ 
    sick_scan_api: struct SickScanRadarObject, equivalent to RadarObject.msg

    Attributes
    ----------
    id : ctypes.c_int32
    tracking_time_sec : ctypes.c_uint32
        seconds part of tracking_time (since when the object is tracked): seconds (stamp_secs) since epoch
    tracking_time_nsec : ctypes.c_uint32
        nanoseconds part of tracking_time (since when the object is tracked): nanoseconds since stamp_secs
    last_seen_sec : ctypes.c_uint32
        seconds part of last_seen timestamp: seconds (stamp_secs) since epoch
    last_seen_nsec : ctypes.c_uint32
        nanoseconds part of last_seen timestamp: nanoseconds since stamp_secs
    # geometry_msgs/TwistWithCovariance velocity
    velocity_linear : SickScanVector3Msg
    velocity_angular : SickScanVector3Msg
    velocity_covariance : ctypes.c_double * 36
        Row-major representation of the 6x6 covariance matrix (x, y, z, rotation about X axis, rotation about Y axis, rotation about Z axis)
    # geometry_msgs/Pose bounding_box_center
    bounding_box_center_position : SickScanVector3Msg
    bounding_box_center_orientation : SickScanQuaternionMsg
    geometry_msgs/Vector3 bounding_box_size
    bounding_box_size : SickScanVector3Msg
    # geometry_msgs/PoseWithCovariance object_box_center
    object_box_center_position : SickScanVector3Msg
    object_box_center_orientation : SickScanQuaternionMsg
    object_box_center_covariance : ctypes.c_double * 36
        Row-major representation of the 6x6 covariance matrix (x, y, z, rotation about X axis, rotation about Y axis, rotation about Z axis)
    # geometry_msgs/Vector3 object_box_size
    object_box_size : SickScanVector3Msg
    geometry_msgs/Point[] contour_points
    contour_points : SickScanPointArray)
    """
    _fields_ = [
        ("id", ctypes.c_int32),
        ("tracking_time_sec", ctypes.c_uint32),      # seconds part of tracking_time (since when the object is tracked): seconds (stamp_secs) since epoch
        ("tracking_time_nsec", ctypes.c_uint32),     # nanoseconds part of tracking_time (since when the object is tracked): nanoseconds since stamp_secs
        ("last_seen_sec", ctypes.c_uint32),          # seconds part of last_seen timestamp: seconds (stamp_secs) since epoch
        ("last_seen_nsec", ctypes.c_uint32),         # nanoseconds part of last_seen timestamp: nanoseconds since stamp_secs
        # geometry_msgs/TwistWithCovariance velocity
        ("velocity_linear", SickScanVector3Msg),
        ("velocity_angular", SickScanVector3Msg),
        ("velocity_covariance", ctypes.c_double * 36), # Row-major representation of the 6x6 covariance matrix (x, y, z, rotation about X axis, rotation about Y axis, rotation about Z axis)
        # geometry_msgs/Pose bounding_box_center
        ("bounding_box_center_position", SickScanVector3Msg),
        ("bounding_box_center_orientation", SickScanQuaternionMsg),
        # geometry_msgs/Vector3 bounding_box_size
        ("bounding_box_size", SickScanVector3Msg),
        # geometry_msgs/PoseWithCovariance object_box_center
        ("object_box_center_position", SickScanVector3Msg),
        ("object_box_center_orientation", SickScanQuaternionMsg),
        ("object_box_center_covariance", ctypes.c_double * 36), # Row-major representation of the 6x6 covariance matrix (x, y, z, rotation about X axis, rotation about Y axis, rotation about Z axis)
        # geometry_msgs/Vector3 object_box_size
        ("object_box_size", SickScanVector3Msg),
        # geometry_msgs/Point[] contour_points
        ("contour_points", SickScanPointArray)
    ]

class SickScanRadarObjectArray(ctypes.Structure):
    """ 
    sick_scan_api: struct SickScanRadarObjectArray, Array of SickScanRadarObject

    Attributes
    ----------
    capacity : ctypes.c_uint64
        Number of allocated elements, i.e. max. number of elements in buffer, allocated buffer size is capacity * sizeof(SickScanRadarObject)
    size : ctypes.c_uint64
        Number of currently used elements in the buffer
    buffer : ctypes.POINTER(SickScanRadarObject)
        Memory, data in plain order and system endianess (buffer == 0, if size == 0 && capacity == 0, otherwise allocated memory), allocation/deallocation always managed by the caller.
    """
    _fields_ = [
        ("capacity", ctypes.c_uint64),                  # Number of allocated elements, i.e. max. number of elements in buffer, allocated buffer size is capacity * sizeof(SickScanRadarObject)
        ("size", ctypes.c_uint64),                      # Number of currently used elements in the buffer
        ("buffer", ctypes.POINTER(SickScanRadarObject)) # Memory, data in plain order and system endianess (buffer == 0, if size == 0 && capacity == 0, otherwise allocated memory), allocation/deallocation always managed by the caller.
    ]

class SickScanRadarScan(ctypes.Structure):
    """ 
    sick_scan_api: struct SickScanRadarScan, equivalent to RadarScan.msg

    Attributes
    ----------
    header : SickScanHeader
        message timestamp
    radarpreheader : SickScanRadarPreHeader
        RadarPreHeader.msg
    targets : SickScanPointCloudMsg
        sensor_msgs/PointCloud2
    objects : SickScanRadarObjectArray
        Array of RadarObject.msg
    """
    _fields_ = [
        ("header", SickScanHeader),                  # message timestamp
        ("radarpreheader", SickScanRadarPreHeader),  # RadarPreHeader.msg
        ("targets", SickScanPointCloudMsg),          # sensor_msgs/PointCloud2
        ("objects", SickScanRadarObjectArray)        # Array of RadarObject.msg
    ]

class SickScanLdmrsObject(SickScanRadarObject):
    """ 
    sick_scan_api: struct SickScanLdmrsObject, equivalent to SickLdmrsObject.msg
    """
    pass

class SickScanLdmrsObjectBuffer(ctypes.Structure):
    """ 
    sick_scan_api: struct SickScanLdmrsObjectBuffer, Array of SickScanLdmrsObject

    Attributes
    ----------
    capacity : ctypes.c_uint64
        Number of allocated elements, i.e. max. number of elements in buffer, allocated buffer size is capacity * sizeof(SickScanLdmrsObject)
    size : ctypes.c_uint64
        Number of currently used elements in the buffer
    buffer : ctypes.POINTER(SickScanLdmrsObject)
        Memory, data in plain order and system endianess (buffer == 0, if size == 0 && capacity == 0, otherwise allocated memory), allocation/deallocation always managed by the caller.
    """
    _fields_ = [
        ("capacity", ctypes.c_uint64),                  # Number of allocated elements, i.e. max. number of elements in buffer, allocated buffer size is capacity * sizeof(SickScanLdmrsObject)
        ("size", ctypes.c_uint64),                      # Number of currently used elements in the buffer
        ("buffer", ctypes.POINTER(SickScanLdmrsObject)) # Memory, data in plain order and system endianess (buffer == 0, if size == 0 && capacity == 0, otherwise allocated memory), allocation/deallocation always managed by the caller.
    ]

class SickScanLdmrsObjectArray(ctypes.Structure):
    """ 
    sick_scan_api: struct SickScanLdmrsObjectArray, equivalent to SickLdmrsObjectArray.msg

    Attributes
    ----------
    header : SickScanHeader
        message timestamp
    objects : SickScanLdmrsObjectBuffer
        Array of SickScanLdmrsObjects
    """
    _fields_ = [
        ("header", SickScanHeader),                    # message timestamp
        ("objects", SickScanLdmrsObjectBuffer)         # Array of SickScanLdmrsObjects
    ]

class SickScanColorRGBA(ctypes.Structure):
    """ 
    equivalent to std_msgs::ColorRGBA

    Attributes
    ----------
    r : ctypes.c_float
    g : ctypes.c_float
    b : ctypes.c_float
    a : ctypes.c_float
    """
    _fields_ = [
        ("r", ctypes.c_float),
        ("g", ctypes.c_float),
        ("b", ctypes.c_float),
        ("a", ctypes.c_float)
    ]

class SickScanColorRGBAArray(ctypes.Structure):
    """ 
    Array of SickScanColorRGBA, which can be serialized and imported in C, C++ or python

    Attributes
    ----------
    capacity : ctypes.c_uint64
        Number of allocated elements, i.e. max. number of elements in buffer, allocated buffer size is capacity * sizeof(SickScanColorRGBA)
    size : ctypes.c_uint64
        Number of currently used elements in the buffer
    buffer : ctypes.POINTER(SickScanColorRGBA)
        Memory, data in plain order and system endianess (buffer == 0, if size == 0 && capacity == 0, otherwise allocated memory), allocation/deallocation always managed by the caller.
    """
    _fields_ = [
        ("capacity", ctypes.c_uint64),                   # Number of allocated elements, i.e. max. number of elements in buffer, allocated buffer size is capacity * sizeof(SickScanColorRGBA)
        ("size", ctypes.c_uint64),                       # Number of currently used elements in the buffer
        ("buffer", ctypes.POINTER(SickScanColorRGBA))    # Memory, data in plain order and system endianess (buffer == 0, if size == 0 && capacity == 0, otherwise allocated memory), allocation/deallocation always managed by the caller.
    ]

class SickScanVisualizationMarker(ctypes.Structure):
    """ 
    equivalent to visualization_msgs::Marker

    Attributes
    ----------
    header : SickScanHeader
        message timestamp
    ns : ctypes.c_char * 1024
        Namespace to place this object in... used in conjunction with id to create a unique name for the object
    id : ctypes.c_int32
        object ID useful in conjunction with the namespace for manipulating and deleting the object later
    type : ctypes.c_int32
        Type of object
    action : ctypes.c_int32
        0 add/modify an object, 1 (deprecated), 2 deletes an object, 3 deletes all objects
    pose_position : SickScanVector3Msg
        Pose of the object (positional part)
    pose_orientation : SickScanQuaternionMsg
        Pose of the object (rotational part)
    scale : SickScanVector3Msg
        Scale of the object 1,1,1 means default (usually 1 meter square)
    color : SickScanColorRGBA
        Color [0.0-1.0]
    lifetime_sec : ctypes.c_uint32
        How long the object should last before being automatically deleted.  0 means forever (seconds part)
    lifetime_nsec : ctypes.c_uint32
        How long the object should last before being automatically deleted.  0 means forever (nanoseconds part)
    frame_locked : ctypes.c_uint8
        boolean, If this marker should be frame-locked, i.e. retransformed into its frame every timestep
    points : SickScanPointArray
        Only used if the type specified has some use for them (eg. POINTS, LINE_STRIP, ...)
    colors : SickScanColorRGBAArray
        Only used if the type specified has some use for them (eg. POINTS, LINE_STRIP, ...). Number of colors must either be 0 or equal to the number of points. NOTE: alpha is not yet used
    text : ctypes.c_char * 1024
        NOTE: only used for text markers
    mesh_resource : ctypes.c_char * 1024
        NOTE: only used for MESH_RESOURCE markers
    mesh_use_embedded_materials : ctypes.c_uint8
        boolean
    """
    _fields_ = [
        ("header", SickScanHeader),                      # message timestamp
        ("ns", ctypes.c_char * 1024),                    # Namespace to place this object in... used in conjunction with id to create a unique name for the object
        ("id", ctypes.c_int32),                          # object ID useful in conjunction with the namespace for manipulating and deleting the object later
        ("type", ctypes.c_int32),                        # Type of object
        ("action", ctypes.c_int32),                      # 0 add/modify an object, 1 (deprecated), 2 deletes an object, 3 deletes all objects
        ("pose_position", SickScanVector3Msg),           # Pose of the object (positional part)
        ("pose_orientation", SickScanQuaternionMsg),     # Pose of the object (rotational part)
        ("scale", SickScanVector3Msg),                   # Scale of the object 1,1,1 means default (usually 1 meter square)
        ("color", SickScanColorRGBA),                    # Color [0.0-1.0]
        ("lifetime_sec", ctypes.c_uint32),               # How long the object should last before being automatically deleted.  0 means forever (seconds part)
        ("lifetime_nsec", ctypes.c_uint32),              # How long the object should last before being automatically deleted.  0 means forever (nanoseconds part)
        ("frame_locked", ctypes.c_uint8),                # boolean, If this marker should be frame-locked, i.e. retransformed into its frame every timestep
        ("points", SickScanPointArray),                  # Only used if the type specified has some use for them (eg. POINTS, LINE_STRIP, ...)
        ("colors", SickScanColorRGBAArray),              # Only used if the type specified has some use for them (eg. POINTS, LINE_STRIP, ...). Number of colors must either be 0 or equal to the number of points. NOTE: alpha is not yet used
        ("text", ctypes.c_char * 1024),                  # NOTE: only used for text markers
        ("mesh_resource", ctypes.c_char * 1024),         # NOTE: only used for MESH_RESOURCE markers
        ("mesh_use_embedded_materials", ctypes.c_uint8)  # boolean
    ]

class SickScanVisualizationMarkerBuffer(ctypes.Structure):
    """
    Array of SickScanVisualizationMarkers

    Attributes
    ----------
    capacity : ctypes.c_uint64
        Number of allocated elements, i.e. max. number of elements in buffer, allocated buffer size is capacity * sizeof(SickScanVisualizationMarker)
    size : ctypes.c_uint64
        Number of currently used elements in the buffer
    buffer : ctypes.POINTER(SickScanVisualizationMarker)
        Memory, data in plain order and system endianess (buffer == 0, if size == 0 && capacity == 0, otherwise allocated memory), allocation/deallocation always managed by the caller.
    """
    _fields_ = [
        ("capacity", ctypes.c_uint64),                          # Number of allocated elements, i.e. max. number of elements in buffer, allocated buffer size is capacity * sizeof(SickScanVisualizationMarker)
        ("size", ctypes.c_uint64),                              # Number of currently used elements in the buffer
        ("buffer", ctypes.POINTER(SickScanVisualizationMarker)) # Memory, data in plain order and system endianess (buffer == 0, if size == 0 && capacity == 0, otherwise allocated memory), allocation/deallocation always managed by the caller.
    ]

class SickScanVisualizationMarkerMsg(ctypes.Structure):
    """ 
    equivalent to visualization_msgs::MarkerArray

    Attributes
    ----------
    markers : SickScanVisualizationMarkerBuffer
        Array of SickScanVisualizationMarkers
    """
    _fields_ = [
        ("markers", SickScanVisualizationMarkerBuffer)          # Array of SickScanVisualizationMarkers
    ]

class SickScanNavReflector(ctypes.Structure):
    """ 
    NAV-350 reflector equivalent to SickScanNavReflector defined in sick_scan_api.h
    """
    _fields_ = [
        ("pos_valid", ctypes.c_uint16),
        # reflector position in [m] in ros coordinates, if pos_valid > 0:
        ("pos_x", ctypes.c_float),
        ("pos_y", ctypes.c_float),
        ("cartesian_valid", ctypes.c_uint16),
        # reflector position in [mm] in lidar coordinates, if cartesian_valid > 0:
        ("cartesian_x", ctypes.c_int32),
        ("cartesian_y", ctypes.c_int32),
        ("polar_valid", ctypes.c_uint16),
        # reflector position in [mm] and [mdeg] in polar lidar coordinates, if polar_valid > 0:
        ("polar_dist", ctypes.c_uint32),
        ("polar_phi", ctypes.c_uint32),
        ("opt_valid", ctypes.c_uint16),
        # Optional reflector data, if opt_valid > 0
        ("opt_local_id", ctypes.c_uint16),
        ("opt_global_id", ctypes.c_uint16),
        ("opt_type", ctypes.c_uint8),
        ("opt_subtype", ctypes.c_uint16),
        ("opt_quality", ctypes.c_uint16),
        ("opt_timestamp", ctypes.c_uint32), # lidar timestamp in milliseconds
        ("opt_size", ctypes.c_uint16),
        ("opt_hitcount", ctypes.c_uint16),
        ("opt_meanecho", ctypes.c_uint16),
        ("opt_startindex", ctypes.c_uint16),
        ("opt_endindex", ctypes.c_uint16),
        ("opt_timestamp_sec", ctypes.c_uint32), # timestamp converted to system time (seconds part, 0 if timestamp not valid)
        ("opt_timestamp_nsec", ctypes.c_uint32) # timestamp converted to system time (nanoseconds part, 0 if timestamp not valid)
    ]

class SickScanNavReflectorBuffer(ctypes.Structure):
    """ 
    Array of NAV-350 reflectors equivalent to SickScanNavReflectorBuffer defined in sick_scan_api.h

    Attributes
    ----------
    capacity : ctypes.c_uint64
        Number of allocated elements, i.e. max. number of elements in buffer, allocated buffer size is capacity * sizeof(SickScanNavReflector)
    size : ctypes.c_uint64
        Number of currently used elements in the buffer
    buffer : ctypes.POINTER(SickScanNavReflector)
        Memory, data in plain order and system endianess (buffer == 0, if size == 0 && capacity == 0, otherwise allocated memory), allocation/deallocation always managed by the caller.
    """
    _fields_ = [
        ("capacity", ctypes.c_uint64),                   # Number of allocated elements, i.e. max. number of elements in buffer, allocated buffer size is capacity * sizeof(SickScanVisualizationMarker)
        ("size", ctypes.c_uint64),                       # Number of currently used elements in the buffer
        ("buffer", ctypes.POINTER(SickScanNavReflector)) # Memory, data in plain order and system endianess (buffer == 0, if size == 0 && capacity == 0, otherwise allocated memory), allocation/deallocation always managed by the caller.
    ]

class SickScanNavPoseLandmarkMsg(ctypes.Structure):
    """ 
    NAV-350 pose and landmark message equivalent to SickScanNavPoseLandmarkMsg defined in sick_scan_api.h
    """
    _fields_ = [
        ("pose_valid", ctypes.c_uint16),
        # NAV pose, if pose_valid > 0:
        ("pose_x", ctypes.c_float),   # x-position in ros coordinates in m
        ("pose_y", ctypes.c_float),   # y-position in ros coordinates in m
        ("pose_yaw", ctypes.c_float), # yaw angle in ros coordinates in radians
        ("pose_timestamp_sec", ctypes.c_uint32), # timestamp of pose converted to system time (seconds part, 0 if timestamp not valid)
        ("pose_timestamp_nsec", ctypes.c_uint32), # timestamp of pose converted to system time (nanoseconds part, 0 if timestamp not valid)
        ("pose_nav_x", ctypes.c_int32), # x-position in lidar coordinates in mm
        ("pose_nav_y", ctypes.c_int32), # y-position in lidar coordinates in mm
        ("pose_nav_phi", ctypes.c_uint32), # orientation in lidar coordinates in 0 ... 360000 mdeg
        ("pose_opt_valid", ctypes.c_uint16),
        # Optional NAV pose data, if pose_opt_valid > 0:
        ("pose_opt_output_mode", ctypes.c_uint8),
        ("pose_opt_timestamp", ctypes.c_uint32), # lidar timestamp in milliseconds
        ("pose_opt_mean_dev", ctypes.c_int32),
        ("pose_opt_nav_mode", ctypes.c_uint8),
        ("pose_opt_info_state", ctypes.c_uint32),
        ("pose_opt_quant_used_reflectors", ctypes.c_uint8),
        # NAV reflectors:
        ("reflectors", SickScanNavReflectorBuffer) # Array of SickScanNavReflectors
    ]

class SickScanNavOdomVelocityMsg(ctypes.Structure):
    """ 
    NAV-350 velocity/odometry data in nav coordinates, see NAVOdomVelocity.msg
    """
    _fields_ = [
        ("vel_x", ctypes.c_float),      # x-component of velocity in the coordinate system defined by coordbase (i.e. in lidar coordinate for coordbase=0) in m/s, -32.0 ... +32.0 m/s
        ("vel_y", ctypes.c_float),      # y-component of velocity in the coordinate system defined by coordbase (i.e. in lidar coordinate for coordbase=0) in m/s, -32.0 ... +32.0 m/s
        ("omega", ctypes.c_float),      # angular velocity of the NAV350 in radians/s, -2*PI ... +2*PI rad/s
        ("timestamp", ctypes.c_uint32), # timestamp of the Velocity vector related to the NAV350 clock
        ("coordbase", ctypes.c_uint8)   # coordinate system of the velocity vector (local or global), 0 = local coordinate system of the NAV350, 1 = absolute coordinate system
    ]

class SickScanOdomVelocityMsg(ctypes.Structure):
    """ 
    Velocity/odometry data in ros coordinates
    """
    _fields_ = [
        ("vel_x", ctypes.c_float),          # x-component of velocity in ros coordinates in m/s
        ("vel_y", ctypes.c_float),          # y-component of velocity in ros coordinates in m/s
        ("omega", ctypes.c_float),          # angular velocity in radians/s
        ("timestamp_sec", ctypes.c_uint32), # seconds part of system timestamp of the odometry data
        ("timestamp_nsec", ctypes.c_uint32) # nanoseconds part of system timestamp of the odometry data
    ]

class SickScanLogMsg(ctypes.Structure):
    """ 
    general log message
    """
    _fields_ = [
        ("log_level", ctypes.c_int32),      # log_level defined in ros::console::levels: Info=1, Warn=2, Error=3, Fatal=4
        ("log_message", ctypes.c_char_p)    # log message
    ]

class SickScanDiagnosticMsg(ctypes.Structure):
    """ 
    general log message
    """
    _fields_ = [
        ("status_code", ctypes.c_int32),    # status_code defined in SICK_DIAGNOSTIC_STATUS: OK=0 (normal operation), WARN=1 (warning), ERROR=2 (error, should not occure), INIT=3 (initialization after startup or reconnection), EXIT=4 (sick_scan_xd exiting)
        ("status_message", ctypes.c_char_p) # diagnostic message
    ]

class SickScanApiErrorCodes(Enum): # 
    """ 
    Error codes, return values of SickScanApi-functions

    Attributes
    ----------
    SICK_SCAN_API_SUCCESS = 0, "SICK_SCAN_API_SUCCESS"                 # function executed successfully
    SICK_SCAN_API_ERROR = 1, "SICK_SCAN_API_ERROR"                     # general (unspecified) error
    SICK_SCAN_API_NOT_LOADED = 2, "SICK_SCAN_API_NOT_LOADED"           # sick_scan_xd library not loaded
    SICK_SCAN_API_NOT_INITIALIZED = 3, "SICK_SCAN_API_NOT_INITIALIZED" # API not initialized
    SICK_SCAN_API_NOT_IMPLEMENTED = 4, "SICK_SCAN_API_NOT_IMPLEMENTED" # function not implemented in sick_scan_xd library
    SICK_SCAN_API_TIMEOUT = 5, "SICK_SCAN_API_TIMEOUT"                 # timeout during wait for response
    """
    SICK_SCAN_API_SUCCESS = 0, "SICK_SCAN_API_SUCCESS"                 # function executed successfully
    SICK_SCAN_API_ERROR = 1, "SICK_SCAN_API_ERROR"                     # general (unspecified) error
    SICK_SCAN_API_NOT_LOADED = 2, "SICK_SCAN_API_NOT_LOADED"           # sick_scan_xd library not loaded
    SICK_SCAN_API_NOT_INITIALIZED = 3, "SICK_SCAN_API_NOT_INITIALIZED" # API not initialized
    SICK_SCAN_API_NOT_IMPLEMENTED = 4, "SICK_SCAN_API_NOT_IMPLEMENTED" # function not implemented in sick_scan_xd library
    SICK_SCAN_API_TIMEOUT = 5, "SICK_SCAN_API_TIMEOUT"                 # timeout during wait for response
    def __int__(self):
        return self.value[0]
    def __str__(self):
        return self.value[1]

""" 
Callback declarations
"""

SickScanPointCloudMsgCallback = ctypes.CFUNCTYPE(None, ctypes.c_void_p, ctypes.POINTER(SickScanPointCloudMsg))                 # sick_scan_api.h: typedef void(* SickScanPointCloudMsgCallback)(SickScanApiHandle apiHandle, const SickScanPointCloudMsg* msg);
SickScanImuMsgCallback = ctypes.CFUNCTYPE(None, ctypes.c_void_p, ctypes.POINTER(SickScanImuMsg))                               # sick_scan_api.h: typedef void(* SickScanImuMsgCallback)(SickScanApiHandle apiHandle, const SickScanImuMsg* msg);
SickScanLFErecMsgCallback = ctypes.CFUNCTYPE(None, ctypes.c_void_p, ctypes.POINTER(SickScanLFErecMsg))                         # sick_scan_api.h: typedef void(* SickScanLFErecMsgCallback)(SickScanApiHandle apiHandle, const SickScanLFErecMsg* msg);
SickScanLIDoutputstateMsgCallback = ctypes.CFUNCTYPE(None, ctypes.c_void_p, ctypes.POINTER(SickScanLIDoutputstateMsg))         # sick_scan_api.h: typedef void(* SickScanLIDoutputstateMsgCallback)(SickScanApiHandle apiHandle, const SickScanLIDoutputstateMsg* msg);
SickScanRadarScanCallback = ctypes.CFUNCTYPE(None, ctypes.c_void_p, ctypes.POINTER(SickScanRadarScan))                         # sick_scan_api.h: typedef void(* SickScanRadarScanCallback)(SickScanApiHandle apiHandle, const SickScanRadarScan* msg);
SickScanLdmrsObjectArrayCallback = ctypes.CFUNCTYPE(None, ctypes.c_void_p, ctypes.POINTER(SickScanLdmrsObjectArray))           # sick_scan_api.h: typedef void(* SickScanLdmrsObjectArrayCallback)(SickScanApiHandle apiHandle, const SickScanLdmrsObjectArray* msg);
SickScanVisualizationMarkerCallback = ctypes.CFUNCTYPE(None, ctypes.c_void_p, ctypes.POINTER(SickScanVisualizationMarkerMsg))  # sick_scan_api.h: typedef void(* SickScanVisualizationMarkerCallback)(SickScanApiHandle apiHandle, const SickScanVisualizationMarkerMsg* msg);
SickScanNavPoseLandmarkCallback = ctypes.CFUNCTYPE(None, ctypes.c_void_p, ctypes.POINTER(SickScanNavPoseLandmarkMsg))          # sick_scan_api.h: typedef void(* SickScanNavPoseLandmarkCallback)(SickScanApiHandle apiHandle, const SickScanNavPoseLandmarkMsg* msg);
SickScanLogMsgCallback = ctypes.CFUNCTYPE(None, ctypes.c_void_p, ctypes.POINTER(SickScanLogMsg))                               # sick_scan_api.h: typedef void(* SickScanLogMsgCallback)(SickScanApiHandle apiHandle, const SickScanLogMsg* msg);
SickScanDiagnosticMsgCallback = ctypes.CFUNCTYPE(None, ctypes.c_void_p, ctypes.POINTER(SickScanDiagnosticMsg))                 # sick_scan_api.h: typedef void(* SickScanDiagnosticMsgCallback)(SickScanApiHandle apiHandle, const SickScanDiagnosticMsg* msg); 


""" 
Functions to initialize and close the API and a lidar
""" 

def ctypesCharArrayToString(char_array):
    """ 
    Returns a python string from a zero terminated ctypes char array
    """ 
    return ''.join([chr(i) for i in char_array]).rstrip('\x00')

def loadLibrary(paths, lib_filname):
    """ 
    loads and returns a library, given its filename and a list of folder 
    """ 
    for path in paths:
        filename = path + lib_filname
        if os.path.exists(filename):
            return ctypes.CDLL(filename)
    return ctypes.CDLL(lib_filname)

def SickScanApiLoadLibrary(paths, lib_filname):
    """ 
    Load sick_scan_library and functions
    """ 
    sick_scan_library = loadLibrary(paths, lib_filname)
    # sick_scan_api.h: SickScanApiHandle SickScanApiCreate(int argc, char** argv);
    sick_scan_library.SickScanApiCreate.argtypes = [ctypes.c_int, ctypes.POINTER(ctypes.c_char_p)]
    sick_scan_library.SickScanApiCreate.restype = ctypes.c_void_p
    # sick_scan_api.h: int32_t SickScanApiRelease(SickScanApiHandle apiHandle);
    sick_scan_library.SickScanApiRelease.argtypes = [ctypes.c_void_p]
    sick_scan_library.SickScanApiRelease.restype = ctypes.c_int
    # sick_scan_api.h: int32_t SickScanApiInitByLaunchfile(SickScanApiHandle apiHandle, const char* launchfile_args);
    sick_scan_library.SickScanApiInitByLaunchfile.argtypes = [ctypes.c_void_p, ctypes.c_char_p]
    sick_scan_library.SickScanApiInitByLaunchfile.restype = ctypes.c_int
    # sick_scan_api.h: int32_t SickScanApiInitByCli(SickScanApiHandle apiHandle, int argc, char** argv);
    sick_scan_library.SickScanApiInitByCli.argtypes = [ctypes.c_void_p, ctypes.c_int, ctypes.POINTER(ctypes.c_char_p)]
    sick_scan_library.SickScanApiInitByCli.restype = ctypes.c_int
    # sick_scan_api.h: int32_t SickScanApiClose(SickScanApiHandle apiHandle);
    sick_scan_library.SickScanApiClose.argtypes = [ctypes.c_void_p]
    sick_scan_library.SickScanApiClose.restype = ctypes.c_int
    # sick_scan_api.h: int32_t SickScanApiRegisterPolarPointCloudMsg(SickScanApiHandle apiHandle, SickScanPointCloudMsgCallback callback);
    sick_scan_library.SickScanApiRegisterCartesianPointCloudMsg.argtypes = [ctypes.c_void_p, SickScanPointCloudMsgCallback]
    sick_scan_library.SickScanApiRegisterCartesianPointCloudMsg.restype = ctypes.c_int
    # sick_scan_api.h: int32_t SickScanApiDeregisterCartesianPointCloudMsg(SickScanApiHandle apiHandle, SickScanPointCloudMsgCallback callback);
    sick_scan_library.SickScanApiDeregisterCartesianPointCloudMsg.argtypes = [ctypes.c_void_p, SickScanPointCloudMsgCallback]
    sick_scan_library.SickScanApiDeregisterCartesianPointCloudMsg.restype = ctypes.c_int
    # sick_scan_api.h: int32_t SickScanApiRegisterPolarPointCloudMsg(SickScanApiHandle apiHandle, SickScanPointCloudMsgCallback callback);
    sick_scan_library.SickScanApiRegisterPolarPointCloudMsg.argtypes = [ctypes.c_void_p, SickScanPointCloudMsgCallback]
    sick_scan_library.SickScanApiRegisterPolarPointCloudMsg.restype = ctypes.c_int
    # sick_scan_api.h: int32_t SickScanApiDeregisterPolarPointCloudMsg(SickScanApiHandle apiHandle, SickScanPointCloudMsgCallback callback);
    sick_scan_library.SickScanApiDeregisterPolarPointCloudMsg.argtypes = [ctypes.c_void_p, SickScanPointCloudMsgCallback]
    sick_scan_library.SickScanApiDeregisterPolarPointCloudMsg.restype = ctypes.c_int
    # sick_scan_api.h: int32_t SickScanApiRegisterImuMsg(SickScanApiHandle apiHandle, SickScanImuMsgCallback callback);
    sick_scan_library.SickScanApiRegisterImuMsg.argtypes = [ctypes.c_void_p, SickScanImuMsgCallback]
    sick_scan_library.SickScanApiRegisterImuMsg.restype = ctypes.c_int
    # sick_scan_api.h: int32_t SickScanApiDeregisterImuMsg(SickScanApiHandle apiHandle, SickScanImuMsgCallback callback);
    sick_scan_library.SickScanApiDeregisterImuMsg.argtypes = [ctypes.c_void_p, SickScanImuMsgCallback]
    sick_scan_library.SickScanApiDeregisterImuMsg.restype = ctypes.c_int
    # sick_scan_api.h: int32_t SickScanApiRegisterLFErecMsg(SickScanApiHandle apiHandle, SickScanLFErecMsgCallback callback);
    sick_scan_library.SickScanApiRegisterLFErecMsg.argtypes = [ctypes.c_void_p, SickScanLFErecMsgCallback]
    sick_scan_library.SickScanApiRegisterLFErecMsg.restype = ctypes.c_int
    # sick_scan_api.h: int32_t SickScanApiDeregisterLFErecMsg(SickScanApiHandle apiHandle, SickScanLFErecMsgCallback callback);
    sick_scan_library.SickScanApiDeregisterLFErecMsg.argtypes = [ctypes.c_void_p, SickScanLFErecMsgCallback]
    sick_scan_library.SickScanApiDeregisterLFErecMsg.restype = ctypes.c_int
    # sick_scan_api.h: int32_t SickScanApiRegisterLIDoutputstateMsg(SickScanApiHandle apiHandle, SickScanLIDoutputstateMsgCallback callback);
    sick_scan_library.SickScanApiRegisterLIDoutputstateMsg.argtypes = [ctypes.c_void_p, SickScanLIDoutputstateMsgCallback]
    sick_scan_library.SickScanApiRegisterLIDoutputstateMsg.restype = ctypes.c_int
    # sick_scan_api.h: int32_t SickScanApiDeregisterLIDoutputstateMsg(SickScanApiHandle apiHandle, SickScanLIDoutputstateMsgCallback callback);
    sick_scan_library.SickScanApiDeregisterLIDoutputstateMsg.argtypes = [ctypes.c_void_p, SickScanLIDoutputstateMsgCallback]
    sick_scan_library.SickScanApiDeregisterLIDoutputstateMsg.restype = ctypes.c_int
    # sick_scan_api.h: int32_t SickScanApiRegisterRadarScanMsg(SickScanApiHandle apiHandle, SickScanRadarScanCallback callback);
    sick_scan_library.SickScanApiRegisterRadarScanMsg.argtypes = [ctypes.c_void_p, SickScanRadarScanCallback]
    sick_scan_library.SickScanApiRegisterRadarScanMsg.restype = ctypes.c_int
    # sick_scan_api.h: int32_t SickScanApiDeregisterRadarScanMsg(SickScanApiHandle apiHandle, SickScanRadarScanCallback callback);
    sick_scan_library.SickScanApiDeregisterRadarScanMsg.argtypes = [ctypes.c_void_p, SickScanRadarScanCallback]
    sick_scan_library.SickScanApiDeregisterRadarScanMsg.restype = ctypes.c_int
    # sick_scan_api.h: int32_t SickScanApiRegisterLdmrsObjectArrayMsg(SickScanApiHandle apiHandle, SickScanLdmrsObjectArrayCallback callback);
    sick_scan_library.SickScanApiRegisterLdmrsObjectArrayMsg.argtypes = [ctypes.c_void_p, SickScanLdmrsObjectArrayCallback]
    sick_scan_library.SickScanApiRegisterLdmrsObjectArrayMsg.restype = ctypes.c_int
    # sick_scan_api.h: int32_t SickScanApiDeregisterLdmrsObjectArrayMsg(SickScanApiHandle apiHandle, SickScanLdmrsObjectArrayCallback callback);
    sick_scan_library.SickScanApiDeregisterLdmrsObjectArrayMsg.argtypes = [ctypes.c_void_p, SickScanLdmrsObjectArrayCallback]
    sick_scan_library.SickScanApiDeregisterLdmrsObjectArrayMsg.restype = ctypes.c_int
    # sick_scan_api.h: int32_t SickScanApiRegisterVisualizationMarkerMsg(SickScanApiHandle apiHandle, SickScanVisualizationMarkerCallback callback);
    sick_scan_library.SickScanApiRegisterVisualizationMarkerMsg.argtypes = [ctypes.c_void_p, SickScanVisualizationMarkerCallback]
    sick_scan_library.SickScanApiRegisterVisualizationMarkerMsg.restype = ctypes.c_int
    # sick_scan_api.h: int32_t SickScanApiDeregisterVisualizationMarkerMsg(SickScanApiHandle apiHandle, SickScanVisualizationMarkerCallback callback);
    sick_scan_library.SickScanApiDeregisterVisualizationMarkerMsg.argtypes = [ctypes.c_void_p, SickScanVisualizationMarkerCallback]
    sick_scan_library.SickScanApiDeregisterVisualizationMarkerMsg.restype = ctypes.c_int
    # sick_scan_api.h:  int32_t SickScanApiRegisterDiagnosticMsg(SickScanApiHandle apiHandle, SickScanDiagnosticMsgCallback callback);
    sick_scan_library.SickScanApiRegisterDiagnosticMsg.argtypes = [ctypes.c_void_p, SickScanDiagnosticMsgCallback]
    sick_scan_library.SickScanApiRegisterDiagnosticMsg.restype = ctypes.c_int
    # sick_scan_api.h:  int32_t SickScanApiDeregisterDiagnosticMsg(SickScanApiHandle apiHandle, SickScanDiagnosticMsgCallback callback);
    sick_scan_library.SickScanApiDeregisterDiagnosticMsg.argtypes = [ctypes.c_void_p, SickScanDiagnosticMsgCallback]
    sick_scan_library.SickScanApiDeregisterDiagnosticMsg.restype = ctypes.c_int
    # sick_scan_api.h:  int32_t SickScanApiRegisterLogMsg(SickScanApiHandle apiHandle, SickScanLogMsgCallback callback);
    sick_scan_library.SickScanApiRegisterLogMsg.argtypes = [ctypes.c_void_p, SickScanLogMsgCallback]
    sick_scan_library.SickScanApiRegisterLogMsg.restype = ctypes.c_int
    # sick_scan_api.h:  int32_t SickScanApiDeregisterLogMsg(SickScanApiHandle apiHandle, SickScanLogMsgCallback callback);
    sick_scan_library.SickScanApiDeregisterLogMsg.argtypes = [ctypes.c_void_p, SickScanLogMsgCallback]
    sick_scan_library.SickScanApiDeregisterLogMsg.restype = ctypes.c_int
    # sick_scan_api.h:  int32_t SickScanApiGetStatus(SickScanApiHandle apiHandle, int32_t* status_code, char* message_buffer, int32_t message_buffer_size);
    sick_scan_library.SickScanApiGetStatus.argtypes = [ctypes.c_void_p, ctypes.POINTER(ctypes.c_int32), ctypes.c_char_p, ctypes.c_int32]
    sick_scan_library.SickScanApiGetStatus.restype = ctypes.c_int
    # sick_scan_api.h:  int32_t SickScanApiSendSOPAS(SickScanApiHandle apiHandle, const char* sopas_command, char* sopas_response_buffer, int32_t response_buffer_size);
    sick_scan_library.SickScanApiSendSOPAS.argtypes = [ctypes.c_void_p, ctypes.c_char_p, ctypes.c_char_p, ctypes.c_int32]
    sick_scan_library.SickScanApiSendSOPAS.restype = ctypes.c_int
    # sick_scan_api.h:  int32_t SickScanApiSetVerboseLevel(SickScanApiHandle apiHandle, int32_t verbose_level);
    sick_scan_library.SickScanApiSetVerboseLevel.argtypes = [ctypes.c_void_p, ctypes.c_int32]
    sick_scan_library.SickScanApiSetVerboseLevel.restype = ctypes.c_int
    # sick_scan_api.h:  int32_t SickScanApiGetVerboseLevel(SickScanApiHandle apiHandle);
    sick_scan_library.SickScanApiGetVerboseLevel.argtypes = [ctypes.c_void_p]
    sick_scan_library.SickScanApiGetVerboseLevel.restype = ctypes.c_int
    # sick_scan_api.h: int32_t SickScanApiWaitNextCartesianPointCloudMsg(SickScanApiHandle apiHandle, SickScanPointCloudMsg* msg, double timeout_sec);
    sick_scan_library.SickScanApiWaitNextCartesianPointCloudMsg.argtypes = [ctypes.c_void_p, ctypes.POINTER(SickScanPointCloudMsg), ctypes.c_double]
    sick_scan_library.SickScanApiWaitNextCartesianPointCloudMsg.restype = ctypes.c_int
    # sick_scan_api.h: int32_t SickScanApiWaitNextPolarPointCloudMsg(SickScanApiHandle apiHandle, SickScanPointCloudMsg* msg, double timeout_sec);
    sick_scan_library.SickScanApiWaitNextPolarPointCloudMsg.argtypes = [ctypes.c_void_p, ctypes.POINTER(SickScanPointCloudMsg), ctypes.c_double]
    sick_scan_library.SickScanApiWaitNextPolarPointCloudMsg.restype = ctypes.c_int
    # sick_scan_api.h: int32_t SickScanApiFreePointCloudMsg(SickScanApiHandle apiHandle, SickScanPointCloudMsg* msg);
    sick_scan_library.SickScanApiFreePointCloudMsg.argtypes = [ctypes.c_void_p, ctypes.POINTER(SickScanPointCloudMsg)]
    sick_scan_library.SickScanApiFreePointCloudMsg.restype = ctypes.c_int
    # sick_scan_api.h: int32_t SickScanApiWaitNextImuMsg(SickScanApiHandle apiHandle, SickScanImuMsg* msg, double timeout_sec);
    sick_scan_library.SickScanApiWaitNextImuMsg.argtypes = [ctypes.c_void_p, ctypes.POINTER(SickScanImuMsg), ctypes.c_double]
    sick_scan_library.SickScanApiWaitNextImuMsg.restype = ctypes.c_int
    # sick_scan_api.h: int32_t SickScanApiFreeImuMsg(SickScanApiHandle apiHandle, SickScanImuMsg* msg)
    sick_scan_library.SickScanApiFreeImuMsg.argtypes = [ctypes.c_void_p, ctypes.POINTER(SickScanImuMsg)]
    sick_scan_library.SickScanApiFreeImuMsg.restype = ctypes.c_int
    # sick_scan_api.h: int32_t SickScanApiWaitNextLFErecMsg(SickScanApiHandle apiHandle, SickScanLFErecMsg* msg, double timeout_sec);
    sick_scan_library.SickScanApiWaitNextLFErecMsg.argtypes = [ctypes.c_void_p, ctypes.POINTER(SickScanLFErecMsg), ctypes.c_double]
    sick_scan_library.SickScanApiWaitNextLFErecMsg.restype = ctypes.c_int
    # sick_scan_api.h: int32_t SickScanApiFreeLFErecMsg(SickScanApiHandle apiHandle, SickScanLFErecMsg* msg);
    sick_scan_library.SickScanApiFreeLFErecMsg.argtypes = [ctypes.c_void_p, ctypes.POINTER(SickScanLFErecMsg)]
    sick_scan_library.SickScanApiFreeLFErecMsg.restype = ctypes.c_int
    # sick_scan_api.h: int32_t SickScanApiWaitNextLIDoutputstateMsg(SickScanApiHandle apiHandle, SickScanLIDoutputstateMsg* msg, double timeout_sec);
    sick_scan_library.SickScanApiWaitNextLIDoutputstateMsg.argtypes = [ctypes.c_void_p, ctypes.POINTER(SickScanLIDoutputstateMsg), ctypes.c_double]
    sick_scan_library.SickScanApiWaitNextLIDoutputstateMsg.restype = ctypes.c_int
    # sick_scan_api.h: int32_t SickScanApiFreeLIDoutputstateMsg(SickScanApiHandle apiHandle, SickScanLIDoutputstateMsg* msg);
    sick_scan_library.SickScanApiFreeLIDoutputstateMsg.argtypes = [ctypes.c_void_p, ctypes.POINTER(SickScanLIDoutputstateMsg)]
    sick_scan_library.SickScanApiFreeLIDoutputstateMsg.restype = ctypes.c_int
    # sick_scan_api.h: int32_t SickScanApiWaitNextRadarScanMsg(SickScanApiHandle apiHandle, SickScanRadarScan* msg, double timeout_sec);
    sick_scan_library.SickScanApiWaitNextRadarScanMsg.argtypes = [ctypes.c_void_p, ctypes.POINTER(SickScanRadarScan), ctypes.c_double]
    sick_scan_library.SickScanApiWaitNextRadarScanMsg.restype = ctypes.c_int
    # sick_scan_api.h: int32_t SickScanApiFreeRadarScanMsg(SickScanApiHandle apiHandle, SickScanRadarScan* msg);
    sick_scan_library.SickScanApiFreeRadarScanMsg.argtypes = [ctypes.c_void_p, ctypes.POINTER(SickScanRadarScan)]
    sick_scan_library.SickScanApiFreeRadarScanMsg.restype = ctypes.c_int
    # sick_scan_api.h: int32_t SickScanApiWaitNextLdmrsObjectArrayMsg(SickScanApiHandle apiHandle, SickScanLdmrsObjectArray* msg, double timeout_sec);
    sick_scan_library.SickScanApiWaitNextLdmrsObjectArrayMsg.argtypes = [ctypes.c_void_p, ctypes.POINTER(SickScanLdmrsObjectArray), ctypes.c_double]
    sick_scan_library.SickScanApiWaitNextLdmrsObjectArrayMsg.restype = ctypes.c_int
    # sick_scan_api.h: int32_t SickScanApiFreeLdmrsObjectArrayMsg(SickScanApiHandle apiHandle, SickScanLdmrsObjectArray* msg);
    sick_scan_library.SickScanApiFreeLdmrsObjectArrayMsg.argtypes = [ctypes.c_void_p, ctypes.POINTER(SickScanLdmrsObjectArray)]
    sick_scan_library.SickScanApiFreeLdmrsObjectArrayMsg.restype = ctypes.c_int
    # sick_scan_api.h: int32_t SickScanApiWaitNextVisualizationMarkerMsg(SickScanApiHandle apiHandle, SickScanVisualizationMarkerMsg* msg, double timeout_sec);
    sick_scan_library.SickScanApiWaitNextVisualizationMarkerMsg.argtypes = [ctypes.c_void_p, ctypes.POINTER(SickScanVisualizationMarkerMsg), ctypes.c_double]
    sick_scan_library.SickScanApiWaitNextVisualizationMarkerMsg.restype = ctypes.c_int
    # sick_scan_api.h: int32_t SickScanApiFreeVisualizationMarkerMsg(SickScanApiHandle apiHandle, SickScanVisualizationMarkerMsg* msg);
    sick_scan_library.SickScanApiFreeVisualizationMarkerMsg.argtypes = [ctypes.c_void_p, ctypes.POINTER(SickScanVisualizationMarkerMsg)]
    sick_scan_library.SickScanApiFreeVisualizationMarkerMsg.restype = ctypes.c_int
    # sick_scan_api.h: int32_t SickScanApiRegisterNavPoseLandmarkMsg(SickScanApiHandle apiHandle, SickScanNavPoseLandmarkCallback callback);
    sick_scan_library.SickScanApiRegisterNavPoseLandmarkMsg.argtypes = [ctypes.c_void_p, SickScanNavPoseLandmarkCallback]
    sick_scan_library.SickScanApiRegisterNavPoseLandmarkMsg.restype = ctypes.c_int
    # sick_scan_api.h: int32_t SickScanApiDeregisterNavPoseLandmarkMsg(SickScanApiHandle apiHandle, SickScanNavPoseLandmarkCallback callback);
    sick_scan_library.SickScanApiDeregisterNavPoseLandmarkMsg.argtypes = [ctypes.c_void_p, SickScanNavPoseLandmarkCallback]
    sick_scan_library.SickScanApiDeregisterNavPoseLandmarkMsg.restype = ctypes.c_int
    # sick_scan_api.h: int32_t SickScanApiWaitNextNavPoseLandmarkMsg(SickScanApiHandle apiHandle, SickScanNavPoseLandmarkMsg* msg, double timeout_sec);
    sick_scan_library.SickScanApiWaitNextNavPoseLandmarkMsg.argtypes = [ctypes.c_void_p, ctypes.POINTER(SickScanNavPoseLandmarkMsg), ctypes.c_double]
    sick_scan_library.SickScanApiWaitNextNavPoseLandmarkMsg.restype = ctypes.c_int
    # sick_scan_api.h: int32_t SickScanApiFreeNavPoseLandmarkMsg(SickScanApiHandle apiHandle, SickScanNavPoseLandmarkMsg* msg);
    sick_scan_library.SickScanApiFreeNavPoseLandmarkMsg.argtypes = [ctypes.c_void_p, ctypes.POINTER(SickScanNavPoseLandmarkMsg)]
    sick_scan_library.SickScanApiFreeNavPoseLandmarkMsg.restype = ctypes.c_int
    # sick_scan_api.h: int32_t SickScanApiNavOdomVelocityMsg(SickScanApiHandle apiHandle, SickScanNavOdomVelocityMsg* msg);
    sick_scan_library.SickScanApiNavOdomVelocityMsg.argtypes = [ctypes.c_void_p, ctypes.POINTER(SickScanNavOdomVelocityMsg)]
    sick_scan_library.SickScanApiNavOdomVelocityMsg.restype = ctypes.c_int
    # sick_scan_api.h: int32_t SickScanApiOdomVelocityMsg(SickScanApiHandle apiHandle, SickScanOdomVelocityMsg* msg);
    sick_scan_library.SickScanApiOdomVelocityMsg.argtypes = [ctypes.c_void_p, ctypes.POINTER(SickScanOdomVelocityMsg)]
    sick_scan_library.SickScanApiOdomVelocityMsg.restype = ctypes.c_int
    return sick_scan_library

def SickScanApiUnloadLibrary(sick_scan_library):
    """ 
    Unload sick_scan_xd api library
    """ 
    del sick_scan_library

def SickScanApiCreate(sick_scan_library):
    """ 
    Create an instance of sick_scan_xd api.
    Call SickScanApiInitByLaunchfile or SickScanApiInitByCli to process a lidar.
    """ 
    null_ptr = ctypes.POINTER(ctypes.c_char_p)()
    api_handle = sick_scan_library.SickScanApiCreate(0, null_ptr)
    return api_handle

def SickScanApiRelease(sick_scan_library, api_handle):
    """ 
    Release and free all resources of the api handle; the handle is invalid after SickScanApiRelease
    """ 
    return sick_scan_library.SickScanApiRelease(api_handle)

def SickScanApiInitByLaunchfile(sick_scan_library, api_handle, launchfile_args):
    """ 
    Initialize a lidar by launchfile and start message receiving and processing
    """ 
    return sick_scan_library.SickScanApiInitByLaunchfile(api_handle, ctypes.create_string_buffer(str.encode(launchfile_args)))

def SickScanApiClose(sick_scan_library, api_handle):
    """ 
    Release and free all resources of the api handle; the handle is invalid after SickScanApiRelease
    """ 
    return sick_scan_library.SickScanApiClose(api_handle)

""" 
Registration and deregistration of message callbacks
""" 

def SickScanApiRegisterCartesianPointCloudMsg(sick_scan_library, api_handle, pointcloud_callback):
    """ 
    Register a callback for cartesian PointCloud messages, pointcloud in cartesian coordinates with fields x, y, z, intensity
    """ 
    return sick_scan_library.SickScanApiRegisterCartesianPointCloudMsg(api_handle, pointcloud_callback)

def SickScanApiDeregisterCartesianPointCloudMsg(sick_scan_library, api_handle, pointcloud_callback):
    """ 
    Deregister a callback for cartesian PointCloud messages, pointcloud in cartesian coordinates with fields x, y, z, intensity
    """ 
    return sick_scan_library.SickScanApiDeregisterCartesianPointCloudMsg(api_handle, pointcloud_callback)

def SickScanApiRegisterPolarPointCloudMsg(sick_scan_library, api_handle, pointcloud_callback):
    """ 
    Register a callback for polar PointCloud messages, pointcloud in polar coordinates with fields range, azimuth, elevation, intensity
    """ 
    return sick_scan_library.SickScanApiRegisterPolarPointCloudMsg(api_handle, pointcloud_callback)

def SickScanApiDeregisterPolarPointCloudMsg(sick_scan_library, api_handle, pointcloud_callback):
    """ 
    Deregister a callback for polar PointCloud messages, pointcloud in polar coordinates with fields range, azimuth, elevation, intensity
    """ 
    return sick_scan_library.SickScanApiDeregisterPolarPointCloudMsg(api_handle, pointcloud_callback)

def SickScanApiRegisterImuMsg(sick_scan_library, api_handle, imu_callback):
    """ 
    Register a callback for Imu messages
    """ 
    return sick_scan_library.SickScanApiRegisterImuMsg(api_handle, imu_callback)

def SickScanApiDeregisterImuMsg(sick_scan_library, api_handle, imu_callback):
    """ 
    Deregister a callback for Imu messages
    """ 
    return sick_scan_library.SickScanApiDeregisterImuMsg(api_handle, imu_callback)

def SickScanApiRegisterLFErecMsg(sick_scan_library, api_handle, lferec_callback):
    """ 
    Register a callback for LFErec messages
    """ 
    return sick_scan_library.SickScanApiRegisterLFErecMsg(api_handle, lferec_callback)

def SickScanApiDeregisterLFErecMsg(sick_scan_library, api_handle, lferec_callback):
    """ 
    Deregister a callback for LFErec messages
    """ 
    return sick_scan_library.SickScanApiDeregisterLFErecMsg(api_handle, lferec_callback)

def SickScanApiRegisterLIDoutputstateMsg(sick_scan_library, api_handle, lidoutputstate_callback):
    """ 
    Register a callback for LIDoutputstate messages
    """ 
    return sick_scan_library.SickScanApiRegisterLIDoutputstateMsg(api_handle, lidoutputstate_callback)

def SickScanApiDeregisterLIDoutputstateMsg(sick_scan_library, api_handle, lidoutputstate_callback):
    """ 
    Deregister a callback for LIDoutputstate messages
    """ 
    return sick_scan_library.SickScanApiDeregisterLIDoutputstateMsg(api_handle, lidoutputstate_callback)

def SickScanApiRegisterRadarScanMsg(sick_scan_library, api_handle, radarscan_callback):
    """ 
    Register a callback for RadarScan messages
    """ 
    return sick_scan_library.SickScanApiRegisterRadarScanMsg(api_handle, radarscan_callback)

def SickScanApiDeregisterRadarScanMsg(sick_scan_library, api_handle, radarscan_callback):
    """ 
    Deregister a callback for RadarScan messages
    """ 
    return sick_scan_library.SickScanApiDeregisterRadarScanMsg(api_handle, radarscan_callback)

def SickScanApiRegisterLdmrsObjectArrayMsg(sick_scan_library, api_handle, ldmrsobjectarray_callback):
    """ 
    Register a callback for LdmrsObjectArray messages
    """ 
    return sick_scan_library.SickScanApiRegisterLdmrsObjectArrayMsg(api_handle, ldmrsobjectarray_callback)

def SickScanApiDeregisterLdmrsObjectArrayMsg(sick_scan_library, api_handle, ldmrsobjectarray_callback):
    """ 
    Deregister a callback for LdmrsObjectArray messages
    """ 
    return sick_scan_library.SickScanApiDeregisterLdmrsObjectArrayMsg(api_handle, ldmrsobjectarray_callback)

def SickScanApiRegisterVisualizationMarkerMsg(sick_scan_library, api_handle, ldmrsobjectarray_callback):
    """ 
    Register a callback for VisualizationMarker messages
    """ 
    return sick_scan_library.SickScanApiRegisterVisualizationMarkerMsg(api_handle, ldmrsobjectarray_callback)

def SickScanApiDeregisterVisualizationMarkerMsg(sick_scan_library, api_handle, ldmrsobjectarray_callback):
    """ 
    Deregister a callback for VisualizationMarker messages
    """ 
    return sick_scan_library.SickScanApiDeregisterVisualizationMarkerMsg(api_handle, ldmrsobjectarray_callback)

def SickScanApiRegisterNavPoseLandmarkMsg(sick_scan_library, api_handle, callback):
    """ 
    Register a callback for SickScanNavPoseLandmarkMsg messages
    """ 
    return sick_scan_library.SickScanApiRegisterNavPoseLandmarkMsg(api_handle, callback)

def SickScanApiDeregisterNavPoseLandmarkMsg(sick_scan_library, api_handle, callback):
    """ 
    Deregister a callback for SickScanNavPoseLandmarkMsg messages
    """ 
    return sick_scan_library.SickScanApiDeregisterNavPoseLandmarkMsg(api_handle, callback)

""" 
Diagnostic functions
""" 

def SickScanApiRegisterDiagnosticMsg(sick_scan_library, api_handle, callback):
    """ 
    Register a callback for diagnostic messages (notification in case of changed status, e.g. after errors)
    """ 
    return sick_scan_library.SickScanApiRegisterDiagnosticMsg(api_handle, callback)

def SickScanApiDeregisterDiagnosticMsg(sick_scan_library, api_handle, callback):
    """ 
    Deregister a callback for diagnostic messages (notification in case of changed status, e.g. after errors)
    """ 
    return sick_scan_library.SickScanApiDeregisterDiagnosticMsg(api_handle, callback)

def SickScanApiRegisterLogMsg(sick_scan_library, api_handle, callback):
    """ 
    Register a callback for log messages (all informational and error messages)
    """ 
    return sick_scan_library.SickScanApiRegisterLogMsg(api_handle, callback)

def SickScanApiDeregisterLogMsg(sick_scan_library, api_handle, callback):
    """ 
    Deregister a callback for log messages (all informational and error messages)
    """ 
    return sick_scan_library.SickScanApiDeregisterLogMsg(api_handle, callback)

def SickScanApiGetStatus(sick_scan_library, api_handle, status_code, message_buffer, message_buffer_size):
    """ 
    Query current status and status message
    """ 
    return sick_scan_library.SickScanApiGetStatus(api_handle, status_code, message_buffer, message_buffer_size)

# sopas_string_buffer = {}
def SickScanApiSendSOPAS(sick_scan_library, api_handle, sopas_command, response_buffer_size = 1024):
    """ 
    Sends a SOPAS command like "sRN SCdevicestate" or "sRN ContaminationResult" and returns the lidar response
    """ 
    response_buffer_size = max(1024, response_buffer_size)
    ctypes_response_buffer = ctypes.create_string_buffer(response_buffer_size + 1)
    ret_val = sick_scan_library.SickScanApiSendSOPAS(api_handle, ctypes.create_string_buffer(str.encode(sopas_command)), ctypes_response_buffer, response_buffer_size)
    # global sopas_string_buffer
    # if sopas_command not in sopas_string_buffer:
    #     sopas_string_buffer[sopas_command] = ctypes.create_string_buffer(str.encode(sopas_command))
    # ret_val = sick_scan_library.SickScanApiSendSOPAS(api_handle, sopas_string_buffer[sopas_command], ctypes_response_buffer, response_buffer_size)
    return ctypes_response_buffer.value.decode()

def SickScanApiSetVerboseLevel(sick_scan_library, api_handle, verbose_level):
    """ 
    Set verbose level 0=DEBUG, 1=INFO, 2=WARN, 3=ERROR, 4=FATAL or 5=QUIET (equivalent to ros::console::levels),
    i.e. print messages on console above the given verbose level.
    Default verbose level is 1 (INFO), i.e. print informational, warnings and error messages.
    """ 
    return sick_scan_library.SickScanApiSetVerboseLevel(api_handle, verbose_level)

def SickScanApiGetVerboseLevel(sick_scan_library, api_handle):
    """ 
    Returns the current verbose level 0=DEBUG, 1=INFO, 2=WARN, 3=ERROR, 4=FATAL or 5=QUIET. Default verbose level is 1 (INFO)
    """ 
    return sick_scan_library.SickScanApiGetVerboseLevel(api_handle)

""" 
Polling functions
""" 

def SickScanApiWaitNextCartesianPointCloudMsg(sick_scan_library, api_handle, msg, timeout_sec):
    """ 
    Wait for and return the next cartesian PointCloud message
    """ 
    return sick_scan_library.SickScanApiWaitNextCartesianPointCloudMsg(api_handle, msg, timeout_sec)

def SickScanApiWaitNextPolarPointCloudMsg(sick_scan_library, api_handle, msg, timeout_sec):
    """ 
    Wait for and return the next polar PointCloud message
    """ 
    return sick_scan_library.SickScanApiWaitNextPolarPointCloudMsg(api_handle, msg, timeout_sec)

def SickScanApiFreePointCloudMsg(sick_scan_library, api_handle, msg):
    """ 
    Deallocate a PointCloud message, use after SickScanApiWaitNextCartesianPointCloudMsg resp. SickScanApiWaitNextPolarPointCloudMsg
    """ 
    return sick_scan_library.SickScanApiFreePointCloudMsg(api_handle, msg)

def SickScanApiWaitNextImuMsg(sick_scan_library, api_handle, msg, timeout_sec):
    """ 
    Wait for and return the next Imu message
    """ 
    return sick_scan_library.SickScanApiWaitNextImuMsg(api_handle, msg, timeout_sec)

def SickScanApiFreeImuMsg(sick_scan_library, api_handle, msg):
    """ 
    Deallocate a Imu message, use after SickScanApiWaitNextImuMsg
    """ 
    return sick_scan_library.SickScanApiFreeImuMsg(api_handle, msg)

def SickScanApiWaitNextLFErecMsg(sick_scan_library, api_handle, msg, timeout_sec):
    """ 
    Wait for and return the next LFErec message
    """ 
    return sick_scan_library.SickScanApiWaitNextLFErecMsg(api_handle, msg, timeout_sec)

def SickScanApiFreeLFErecMsg(sick_scan_library, api_handle, msg):
    """ 
    Deallocate a LFErec message, use after SickScanApiWaitNextLFErecMsg
    """ 
    return sick_scan_library.SickScanApiFreeLFErecMsg(api_handle, msg)

def SickScanApiWaitNextLIDoutputstateMsg(sick_scan_library, api_handle, msg, timeout_sec):
    """ 
    Wait for and return the next LIDoutputstate message
    """ 
    return sick_scan_library.SickScanApiWaitNextLIDoutputstateMsg(api_handle, msg, timeout_sec)

def SickScanApiFreeLIDoutputstateMsg(sick_scan_library, api_handle, msg):
    """ 
    Deallocate a LIDoutputstate message, use after SickScanApiWaitNextLIDoutputstateMsg
    """ 
    return sick_scan_library.SickScanApiFreeLIDoutputstateMsg(api_handle, msg)

def SickScanApiWaitNextRadarScanMsg(sick_scan_library, api_handle, msg, timeout_sec):
    """ 
    Wait for and return the next RadarScan message
    """ 
    return sick_scan_library.SickScanApiWaitNextRadarScanMsg(api_handle, msg, timeout_sec)

def SickScanApiFreeRadarScanMsg(sick_scan_library, api_handle, msg):
    """ 
    Deallocate a RadarScan message, use after SickScanApiWaitNextRadarScanMsg
    """ 
    return sick_scan_library.SickScanApiFreeRadarScanMsg(api_handle, msg)

def SickScanApiWaitNextLdmrsObjectArrayMsg(sick_scan_library, api_handle, msg, timeout_sec):
    """ 
    Wait for and return the next LdmrsObjectArray message
    """ 
    return sick_scan_library.SickScanApiWaitNextLdmrsObjectArrayMsg(api_handle, msg, timeout_sec)

def SickScanApiFreeLdmrsObjectArrayMsg(sick_scan_library, api_handle, msg):
    """ 
    Deallocate a LdmrsObjectArray message, use after SickScanApiWaitNextLdmrsObjectArrayMsg
    """ 
    return sick_scan_library.SickScanApiFreeLdmrsObjectArrayMsg(api_handle, msg)

def SickScanApiWaitNextVisualizationMarkerMsg(sick_scan_library, api_handle, msg, timeout_sec):
    """ 
    Wait for and return the next VisualizationMarker message
    """ 
    return sick_scan_library.SickScanApiWaitNextVisualizationMarkerMsg(api_handle, msg, timeout_sec)

def SickScanApiFreeVisualizationMarkerMsg(sick_scan_library, api_handle, msg):
    """ 
    Deallocate a VisualizationMarker message, use after SickScanApiWaitNextVisualizationMarkerMsg
    """ 
    return sick_scan_library.SickScanApiFreeVisualizationMarkerMsg(api_handle, msg)

def SickScanApiWaitNextNavPoseLandmarkMsg(sick_scan_library, api_handle, msg, timeout_sec):
    """ 
    Wait for and return the next SickScanNavPoseLandmarkMsg message
    """ 
    return sick_scan_library.SickScanApiWaitNextNavPoseLandmarkMsg(api_handle, msg, timeout_sec)

def SickScanApiFreeNavPoseLandmarkMsg(sick_scan_library, api_handle, msg):
    """ 
    Deallocate a SickScanNavPoseLandmarkMsg message, use after SickScanApiWaitNextNavPoseLandmarkMsg
    """ 
    return sick_scan_library.SickScanApiFreeNavPoseLandmarkMsg(api_handle, msg)

def SickScanApiNavOdomVelocityMsg(sick_scan_library, api_handle, msg):
    """ 
    Send NAV350 velocity/odometry data in nav coordinates
    """ 
    return sick_scan_library.SickScanApiNavOdomVelocityMsg(api_handle, msg)

def SickScanApiOdomVelocityMsg(sick_scan_library, api_handle, msg):
    """ 
    Send velocity/odometry data in ros coordinates
    """ 
    return sick_scan_library.SickScanApiOdomVelocityMsg(api_handle, msg)



from phenomate_core.preprocessing.base import BasePreprocessor
from phenomate_core.preprocessing.hyperspec.process import HyperspecPreprocessor
from phenomate_core.preprocessing.imu.process import ImuPreprocessor
from phenomate_core.preprocessing.jai.process import JaiPreprocessor
from phenomate_core.preprocessing.lidar2d.process import Lidar2DPreprocessor
from phenomate_core.preprocessing.lidar3douster.process import Ouster3dPreprocessor 
from phenomate_core.preprocessing.rs3basestation.process import RS3Preprocessor
from phenomate_core.preprocessing.canbus.process import CanbusPreprocessor
from phenomate_core.preprocessing.oak_d.process import (
    OakCalibrationPreprocessor,
    OakFramePreprocessor,
    OakImuPacketsPreprocessor,
)

__all__ = (
    "BasePreprocessor",
    "HyperspecPreprocessor",
    "JaiPreprocessor",
    "OakCalibrationPreprocessor",
    "OakFramePreprocessor",
    "OakImuPacketsPreprocessor",
    "Ouster3dPreprocessor",
    "RS3Preprocessor",
    "CanbusPreprocessor",
)

from phenomate_core.get_version import get_task_logger
shared_logger = get_task_logger(__name__)

def get_preprocessor(sensor: str, details: str = "") -> type[BasePreprocessor]:
    """
    Selects the processing dependant on the keyword that is present in the data file's filename.
    
    N.B. The order of the case statements is important - in particular for the 2d / 3d lidar.
    
    The 'sensor' value is extracted from the filename of the data file, using the appm / appn-project-manager code
    as specified with the template.yaml input file.
    
    N.B. This code is run in the Phenomate\backend\activity\tasks.py preprocess_task() function as a Celery task.
  
    """
    shared_logger.info(f"phenomate_core: get_preprocessor() called with sensor: {sensor}, details: {details}")
    match sensor.lower():
        case sensor if "jai" in sensor:
            return JaiPreprocessor
        case sensor if "hyper" in sensor:
            return HyperspecPreprocessor
        case sensor if "white" in sensor:
            return HyperspecPreprocessor
        case sensor if "dark" in sensor:
            return HyperspecPreprocessor
        case sensor if "oak" in sensor:
            if "calibration" in details:
                return OakCalibrationPreprocessor
            if "imu" in details:
                return OakImuPacketsPreprocessor
            return OakFramePreprocessor
        case sensor if "imu" in sensor:
            return ImuPreprocessor
        case sensor if "ouster" in sensor:
            return Ouster3dPreprocessor
        case sensor if "lidar" in sensor:
            return Lidar2DPreprocessor
        case sensor if "rs3" in sensor:
            return RS3Preprocessor
        case sensor if "canbus" in sensor:
            return CanbusPreprocessor
            
    raise ValueError(f"Unsupported sensor type: {sensor}")



from __future__ import annotations

import os
import sys

import struct
from pathlib import Path
from typing import Any

from datetime import datetime, timezone
import time
import csv

import numpy as np
import ctypes
import datatable as dt
import datatable.math as mt

from phenomate_core.preprocessing.base import BasePreprocessor
from phenomate_core.preprocessing.lidar2d import lidar_pb2
from  phenomate_core.preprocessing.lidar2d.sick_scan_api import ctypesCharArrayToString
from  phenomate_core.preprocessing.lidar2d.reading_proto_buff import from_proto

# This can be used to get parallel processing in the datatable library
# It should be put in the .env and .env.production file
dt.options.nthreads = 1 

from phenomate_core.get_version import get_task_logger
shared_logger = get_task_logger(__name__)

from phenomate_core.get_version import get_version

import psutil 

def check_memory_usage(notes: str):
    memory = psutil.virtual_memory()
    shared_logger.info(f"{notes}: Available Memory: {memory.available / (1024 ** 3):.4f} GB; Used Memory: {memory.used / (1024 ** 3):.4f} GB; Memory Usage: {memory.percent}%")




class Lidar2DPreprocessor(BasePreprocessor[lidar_pb2.SickScanPointCloudMsg]):
    """
    lidar_pb2.SickScanPointCloudMsg is the self.images list type
    
    Procedure edited from the Resonate resonatesystems-rs24005-appn-instrument-interfaces git (Bitbucket) 
    project process_lidar_binaryfiles.py file.
    
    3 input files are tested for speed of data conversion, writing to CSV, memory footprint and final data size.
    Times and sizes are linear with input data size.
    
    | Input bin / MB | messages       | elapsed time / s  |  <--- this is the time range from the 1st and last timestamp
    |----------------|----------------|-------------------|
    |      49.8      |   3670         |     6.115         |
    |      121       |   8938         |     14.89         |
    |      488       |   36000        |     60.0          |
     
    Original using Python lists
    | Input bin / MB | output CSV / MB | MemDiff / GB  | Processing Time / sec | CSV write / sec |
    |----------------|-----------------|---------------|-----------------------|-----------------|
    |      49.8      |   209           |       0.66    |   13.19               |     11.07       |     
    |      121       |   509           |       1.29    |   33.18               |     28.64       |  
    |      488       |   2051          |       5.34    |   125.59              |     112.97      | 
                                                        
    Numpy pre-allocated                                 
    | Input bin / MB | output CSV / MB | MemDiff / GB  | Processing Time / sec | CSV write / sec |
    |----------------|-----------------|---------------|-----------------------|-----------------|
    |      49.8      |   212           |       0.245   |   6.10                |     6.67        |     
    |      121       |   517           |       0.469   |   14.04               |     14.41       |  
    |      488       |   2084          |       1.755   |   54.57               |     57.65       |
                                       
    Numpy pre-allocated convert to 
    datatable            
    | Input bin / MB | output CSV / MB | MemDiff / GB  | Processing Time / sec | CSV write / sec |CSV write (4 threads) / sec |
    |----------------|-----------------|---------------|-----------------------|-----------------|----------------------------|
    |      49.8      |   204           |      0.208    |       6.50            |     1.016       |                            |
    |      121       |   497           |      0.391    |       16.6            |     2.316       |                            |
    |      488       |   2001          |      1.777    |       55.33           |     8.546       |                            |
    |      49.8      |   204           |      0.208    |       7.49            |                 |         0.350              |
    |      121       |   497           |      0.400    |       15.08           |                 |         0.774              |  
    |      488       |   2001          |      1.688    |       56.76           |                 |         3.236              |
    
    Numpy pre-allocated pivoted row per
    message - convert to datatable 
    (6 decimal places)          
    | Input bin / MB | output CSV / MB | MemDiff / GB  | Processing Time / sec | CSV write / sec |
    |----------------|-----------------|---------------|-----------------------|-----------------|
    |      49.8      |   85.3          |      0.106    |       8.03            |     0.475       |
    |      121       |   207           |      0.391    |       15.50           |     1.03        |
    |      488       |   837           |      1.281    |       54.17           |     5.86        |
    
    Numpy pre-allocated, ctype buffer, 
    pivoted row per message - convert 
    to datatable (6 decimal places)    
    | Input bin / MB | output CSV / MB | MemDiff / GB  | Processing Time / sec | CSV write / sec |
    |----------------|-----------------|---------------|-----------------------|-----------------|
    |      49.8      |   85.3          |      0.186    |       0.163           |     0.591       |
    |      121       |   207           |      0.383    |       0.368           |     1.74        |
    |      488       |   837           |      1.310    |       1.421           |     6.388       |
    """
    
    def __init__(self,  path: str | Path, in_ext: str = "bin"):
        super().__init__(path, in_ext)  # Call the parent class constructor
        self.total_messages = 0     # Add subclass-specific data
        self.msg_timestamp = []
        self.msg_height = 0
        self.msg_width = 0
        self.msg_num_fields = 0
        self.msg_data_size = 0
        self.msg_data_buffer = 0
        self.total_points_x = []
        self.total_points_y = []
        self.total_points_z = []
        self.total_timestamps = [] # For Amiga timestamps # use : self.system_timestamps
        self.total_lidar_timestamps = [] # For SickScan LIDAR timestamp
        self.total_points_intensity = []
        self.filtered_data = False
        self.total_xyzi = None
        self.row_offset = 0
        self.total_z_sum = 0.0
        
    def extract(self, **kwargs: Any) -> None:
        check_memory_usage('extract 1')
        with self.path.open("rb") as file:
            
            while True:
                # Read the length of the next serialized message
                serialized_timestamp = file.read(8)
                if not serialized_timestamp:
                    break
                system_timestamp = struct.unpack("d", serialized_timestamp)[0]

                length_bytes = file.read(4)
                if not length_bytes:
                    break
                message_length = int.from_bytes(length_bytes, byteorder="little")

                # Read the serialized message
                serialized_lidar_msg = file.read(message_length)

                # Parse the protobuf message
                protbuf_msg = lidar_pb2.SickScanPointCloudMsg()
                
                try:
                    protbuf_msg.ParseFromString(serialized_lidar_msg)
                    self.msg_timestamp.append(protbuf_msg.header.timestamp_sec + protbuf_msg.header.timestamp_nsec / 1e9)
                    self.system_timestamps.append(system_timestamp)
                    # Update to extracted image list
                    self.images.append(protbuf_msg)
                    self.total_messages += 1
                    
                except Exception as ex:
                    print(f"Error processing message: {ex}")
                    continue
                
                
                # shared_logger.info(f"Converted timestamp: system_timestamp:{system_timestamp}")
                
        # check_memory_usage('extract 2')
            
    def save(
        self,
        path: Path | str,
        width: int | None = None,
        height: int | None = None,
        **kwargs: Any,
    ) -> None:
        """
        This save() extracts the x,y,z,i data for each timestamp, create a row for each timestamp consisting of
        interleaved x,y,z,i values - e.g. 
        <amiga_timestamp>,<instrument_ts>, x0,y0,z0,i0, x1,y1,z1,i1,... xN,yN,zN,iN
        where N = 
        and saves the data to CSV using the datatable table to_csv() method, which can be run wih 
        multiple threads for improved performance (currently set to: dt.options.nthreads = 1 )
        """
        fpath = Path(path)
        fpath.mkdir(parents=True, exist_ok=True)
        
        # Some extra metadata which is not currently used
        current_year = str(datetime.now().year)
        phenomate_version = get_version()        
        user = "Phenomate user"
        
        
        start_time = time.time()
        for index, sickscan_lidar_protobuf_obj in enumerate(self.images):
            
            processed_msg = from_proto(sickscan_lidar_protobuf_obj)

            xyzi_res = self.py_sick_scan_cartesian_point_cloud_msg_to_xy_numpy(
                                processed_msg, index
                )
            points_xyzi = xyzi_res.reshape(( xyzi_res.shape[0] * 4))
            Nreps = xyzi_res.shape[0]
            
            # On reading the first data result, compute the shape 
            # of the final data
            if index == 0:
                try:
                    shared_logger.info(f"LIDAR SickScan Processing: Allocating numpy array of shape: ({self.total_messages}, {2 + (xyzi_res.shape[0] * 4)})")
                    self.total_xyzi= np.zeros((self.total_messages, 2 + (xyzi_res.shape[0] * 4)))
                except MemoryError as ex:
                    shared_logger.error(f"LIDAR SickScan Processing: Error allocating total_xyzi array message: {ex}")
                
            
            if xyzi_res is not None:
                # sum of the Z column, used to check for required 
                # formating when saving to csv - if 0.0 then do not 
                # add extra digits past the decimal
                self.total_z_sum += np.sum(xyzi_res[:, 2])  
                # Add the row of data t the numpy array
                self.total_xyzi[index, 2:] = points_xyzi               
                self.total_xyzi[index, 0:2] = np.array([self.system_timestamps[index],self.msg_timestamp[index]])
  
        end_time = time.time()
        # Print elapsed time
        total_points = self.total_xyzi.shape[0]
        shared_logger.info(f"LIDAR SickScan Processing: Total messages: {self.total_messages} ; Total xyz points: {total_points} ; Points per message: {total_points / self.total_messages}")
        shared_logger.info(f"LIDAR SickScan Processing: Preprocessing time : {end_time - start_time:.4f} seconds")
        check_memory_usage('save 4')
        
        
        start_time = time.time() 
        self.filtered_data = False
        csv_path_name_ext = fpath / self.get_output_name(index = None, ext = "csv", details = "dtpivot_ctype")
        
        
        header_base = [ "X", "Y", "Z", "Intensity"]
        # Nreps is typically 841 for the current 2D Lidar
        header_rep = [f"{col}{i}" for i in range(Nreps) for col in header_base]
        header = ["Amiga_timestamp", "Lidar_timestamp"] + header_rep

        # Convert to datatable Frame
        total_xyzi_data_dt = dt.Frame(self.total_xyzi, names=header)
        
        # Round the x,y,z data to reduce the CSV data size
        for col_name in total_xyzi_data_dt.names:
            total_xyzi_data_dt[col_name] = mt.round(total_xyzi_data_dt[col_name], ndigits=6)

        total_xyzi_data_dt["Amiga_timestamp"] = mt.round(total_xyzi_data_dt["Amiga_timestamp"], ndigits=10)
        total_xyzi_data_dt["Lidar_timestamp"] = mt.round(total_xyzi_data_dt["Lidar_timestamp"], ndigits=10)
        
        dt.options.progress.enabled = False
        total_xyzi_data_dt.to_csv(str(csv_path_name_ext), verbose=False)
            
        end_time = time.time()
        
        shared_logger.info(f"LIDAR SickScan Processing: CSV write time : {end_time - start_time:.4f} seconds")
        check_memory_usage('save 5')  
    
    def matched_file_list(self, origin_path: Path, file_part : str) -> list[Path]:
        """
        Not yet required in this derived class
        """
        shared_logger.info("LIDAR SickScan.copy_extra_files() not implemented")
        return []
                
    def copy_extra_files(self, fpath: Path) -> None:
        """
        Not yet required in this derived class
        """
        shared_logger.info("LIDAR SickScan.copy_extra_files() not implemented")
    
    def save_datatable(
        self,
        path: Path | str,
        width: int | None = None,
        height: int | None = None,
        **kwargs: Any,
    ) -> None:
        """
        This save() extracts the x,y,z,i data using and saves as csv using datatable
        e.g.
        Amiga_timestamp,Lidar_timestamp,X,Y,Z,Intensity
        1755152988.142256,1755152988.1121805,1.2043993,-0.8433294,0.0,249.0
        1755152988.142256,1755152988.1121805,1.213824,-0.8473013,0.0,244.0
        1755152988.142256,1755152988.1121805,1.213331,-0.8443362,0.0,253.0
        """
        fpath = Path(path)
        fpath.mkdir(parents=True, exist_ok=True)
        current_year = str(datetime.now().year)
        phenomate_version = get_version()
        
        # check_memory_usage('extract 3')
        
        user = "Phenomate user" # Creator of the image
        start_time = time.time()
        for index, sickscan_lidar_protobuf_obj in enumerate(self.images):
            
            processed_msg = from_proto(sickscan_lidar_protobuf_obj)

            xyzi_res = self.py_sick_scan_cartesian_point_cloud_msg_to_xy_numpy(
                                processed_msg, index
                )

            # On reading the first data result, compute the shape 
            # of the final data
            if index == 0:
                try:
                    numrows = xyzi_res.shape[0] * self.total_messages
                    shared_logger.info(f"LIDAR SickScan Processing: Allocating numpy array of shape: ({numrows},6)")
                    self.total_xyzi= np.zeros((numrows, 6))
                except MemoryError as ex:
                    shared_logger.error(f"LIDAR SickScan Processing: Error allocating total_xyzi array message: {ex}")
                
            # sum of the Z column, used to check for required 
            # formating when saving to csv - if 0.0 then do not 
            # add extra digits past the decimal
            if xyzi_res is not None:
                self.total_z_sum += np.sum(xyzi_res[:, 2])  

                row_end = self.row_offset + xyzi_res.shape[0]
                # Add the x,y,z,i data
                self.total_xyzi[self.row_offset:row_end, 2:6] = xyzi_res 
                # add the timstamps (repeats the values for each x,y,z,i added from the xyzi_res)
                self.total_xyzi[self.row_offset:row_end, 0:2] = np.array([self.system_timestamps[index],self.msg_timestamp[index]])
                # update the position to the next empty row
                self.row_offset += xyzi_res.shape[0]
        
        end_time = time.time()
        # Print elapsed time
        total_points = self.total_xyzi.shape[0]
        shared_logger.info(f"LIDAR SickScan Processing: Total messages: {self.total_messages} ; Total xyz points: {total_points} ; Points per message: {total_points / self.total_messages}")
        shared_logger.info(f"LIDAR SickScan Processing: Preprocessing time : {end_time - start_time:.4f} seconds")
        check_memory_usage('save 4.0')
        
        
        start_time = time.time() 
        self.filtered_data = False
        csv_path_name_ext = fpath / self.get_output_name(index = None, ext = "csv", details = "datatable_ctypeptr")
        
        header = ["Amiga_timestamp", "Lidar_timestamp", "X", "Y", "Z", "Intensity"]
        # Convert to datatable Frame
        total_xyzi_data_dt = dt.Frame(self.total_xyzi, names=header)
        
        total_xyzi_data_dt["Amiga_timestamp"] = mt.round(total_xyzi_data_dt["Amiga_timestamp"], ndigits=9)
        total_xyzi_data_dt["Lidar_timestamp"] = mt.round(total_xyzi_data_dt["Lidar_timestamp"], ndigits=9)
        total_xyzi_data_dt["X"] = mt.round(total_xyzi_data_dt["X"], ndigits=7)
        total_xyzi_data_dt["Y"] = mt.round(total_xyzi_data_dt["Y"], ndigits=7)
        total_xyzi_data_dt["Z"] = mt.round(total_xyzi_data_dt["Z"], ndigits=7)
        
        dt.options.progress.enabled = False
        if self.total_z_sum == 0.0:
            fmt = ('%.9f', '%.9f',  '%.7f' , '%.7f', '%.1f', '%d')
        else:
            fmt = ('%.9f', '%.9f',  '%.7f' , '%.7f', '%.7f', '%d')
        total_xyzi_data_dt.to_csv(str(csv_path_name_ext), verbose=False)
            
        end_time = time.time()
        # shared_logger.info(f"Saving LIDAR data: {image_path_name_ext}  {utc_datetime}")
        shared_logger.info(f"LIDAR SickScan Processing: CSV write time : {end_time - start_time:.4f} seconds")
        check_memory_usage('save 5')  
    
    
    
    def save_numpy(
        self,
        path: Path | str,
        width: int | None = None,
        height: int | None = None,
        **kwargs: Any,
    ) -> None:
        """
        This save() extracts the x,y,z,i data using a 2D Numpy array
        
        Numpy.savetxt() is a little slow
        """
        fpath = Path(path)
        fpath.mkdir(parents=True, exist_ok=True)
        current_year = str(datetime.now().year)
        phenomate_version = get_version()
        
        check_memory_usage('save 0')
        
        user = "Phenomate user" # Creator of the image
        start_time = time.time()
        for index, sickscan_lidar_protobuf_obj in enumerate(self.images):
            
            processed_msg = from_proto(sickscan_lidar_protobuf_obj)

            result = self.py_sick_scan_cartesian_point_cloud_msg_to_xy_numpy(
                                processed_msg, index
                )

            if index == 0:
                try:
                    numrows = result.shape[0] * self.total_messages
                    shared_logger.info(f"LIDAR SickScan Processing: Allocating numpy array of shape: ({numrows},6)")
                    self.total_xyzi= np.zeros((numrows, 6))
                except MemoryError as ex:
                    shared_logger.error(f"LIDAR SickScan Processing: Error allocating total_xyzi array message: {ex}")
                
            
            if result is not None:
                self.total_z_sum += np.sum(result[:, 2])  # sum of the Z column, used to check for required formating when saving to csv
                
                # self.total_xyzi = np.vstack((self.total_xyzi, result))
                row_end = self.row_offset + result.shape[0]
                self.total_xyzi[self.row_offset:row_end, 2:6] = result                
                self.total_xyzi[self.row_offset:row_end, 0:2] = np.array([self.system_timestamps[index],self.msg_timestamp[index]])
                # self.total_xyzi[self.row_offset:row_end, 1] = self.msg_timestamp[index]
                
                self.row_offset += result.shape[0]
        
        end_time = time.time()
        # Print elapsed time
        total_points = len(self.total_points_x)
        shared_logger.info(f"LIDAR SickScan Processing: Total messages: {self.total_messages} ; Total xyz points: {total_points} ; Points per message: {total_points / self.total_messages}")
        shared_logger.info(f"LIDAR SickScan Processing: Preprocessing time : {end_time - start_time:.4f} seconds")
        check_memory_usage('save 1')
        
        
        start_time = time.time() 
        self.filtered_data = False
        csv_path_name_ext = fpath / self.get_output_name(index = None, ext = "csv", details = "numpy")
        
        header = ["Amiga_timestamp", "Lidar_timestamp", "X", "Y", "Z", "Intensity"]
        if self.total_z_sum == 0.0:
            fmt = ('%.9f', '%.9f',  '%.7f' , '%.7f', '%.1f', '%d')
        else:
            fmt = ('%.9f', '%.9f',  '%.7f' , '%.7f', '%.7f', '%d')
        np.savetxt(csv_path_name_ext, self.total_xyzi, delimiter=",", header= ",".join(header), comments='', fmt=fmt)
            
        end_time = time.time()
        # shared_logger.info(f"Saving LIDAR data: {image_path_name_ext}  {utc_datetime}")
        shared_logger.info(f"LIDAR SickScan Processing: CSV write time ): {end_time - start_time:.4f} seconds")
        check_memory_usage('save 2')   
    
    
    def py_sick_scan_cartesian_point_cloud_msg_to_xy_numpy(
        self, pointcloud_msg, index=0, start_time=None
    ):
        """
        This method converts the pointcloud_msg to x,y coordinates.
        
        Taken from the Resonate resonatesystems-rs24005-appn-instrument-interfaces git (Bitbucket) 
        project process_lidar_binaryfiles.py file.
        """

        num_fields = pointcloud_msg.fields.size
        msg_fields_buffer = pointcloud_msg.fields.buffer

        # Initialize offsets to None or some default value
        field_offset_x = -1
        field_offset_y = -1
        field_offset_z = -1
        field_offset_intensity = -1

        for n in range(num_fields):
            field_name = ctypesCharArrayToString(msg_fields_buffer[n].name)
            field_offset = msg_fields_buffer[n].offset
            if field_name == "x":
                field_offset_x = msg_fields_buffer[n].offset
            elif field_name == "y":
                field_offset_y = msg_fields_buffer[n].offset
            elif field_name == "z":
                field_offset_z = msg_fields_buffer[n].offset
            elif field_name == "intensity":
                field_offset_intensity = msg_fields_buffer[n].offset

        if (
            field_offset_x is None
            or field_offset_y is None
            or (field_offset_z is None)
        ):
            raise ValueError("LIDAR SickScan Processing: Offsets not assigned correctly.")

        # row_step = width * 16   (16 == 4 x fp32 value - x, y, z, intensity)
        # This is the length in bytes of the full message array
        cloud_data_buffer_len = pointcloud_msg.row_step * pointcloud_msg.height 


        if ( not (
            pointcloud_msg.data.size == cloud_data_buffer_len
            and num_fields == 4
            and field_offset_x == 0
            and field_offset_y == 4
            and field_offset_z == 8
            and field_offset_intensity == 12             
            )
        ):
            raise ValueError("LIDAR SickScan Processing: Data offsets not as expected.")

        total_floats = pointcloud_msg.width * pointcloud_msg.height * 4
        
        # if index < 1:
            # shared_logger.info(f"pointcloud_msg.width     : {pointcloud_msg.width}")
            # shared_logger.info(f"pointcloud_msg.height    : {pointcloud_msg.height}")
            # shared_logger.info(f"total_floats             : {total_floats}")
            # shared_logger.info(f"pointcloud_msg.row_step  : {pointcloud_msg.row_step} bytes")
            # shared_logger.info(f"Type of data.buffer      : {type(pointcloud_msg.data.buffer)}")
            # shared_logger.info(f"Type of data.buffer[{index}]: {type(pointcloud_msg.data.buffer[index])}")
            # shared_logger.info(f" num_fields {num_fields} ;  field_offset_x {field_offset_x} ; field_offset_y {field_offset_y} ; field_offset_z {field_offset_z} ; field_offset_intensity {field_offset_intensity} ;  width {pointcloud_msg.width} ; height {pointcloud_msg.height};  row_step {pointcloud_msg.row_step} ; point_step {pointcloud_msg.point_step}")
        
        try:
            buffer = ctypes.cast(pointcloud_msg.data.buffer, ctypes.POINTER(ctypes.c_float * total_floats)).contents

            points_xyzi_1d =  np.frombuffer(
                            buffer,
                            dtype=np.float32,
                            count=total_floats,
                            offset=0,
                        )
        except Exception as e:
            shared_logger.error(f"LIDAR SickScan Processing: Error reading buffer for message {index}: {e}")
                    
        # map the read in data to a 2D array            
        points_xyzi = points_xyzi_1d.reshape(( pointcloud_msg.width * pointcloud_msg.height, 4))
        
        return points_xyzi
        
        
    def save_original(
        self,
        path: Path | str,
        width: int | None = None,
        height: int | None = None,
        **kwargs: Any,
    ) -> None:
        """
             
        """
        fpath = Path(path)
        fpath.mkdir(parents=True, exist_ok=True)
        current_year = str(datetime.now().year)
        phenomate_version = get_version()
        
        check_memory_usage('extract 3')
        
        user = "Phenomate user" # Creator of the image
        start_time = time.time()
        for index, sickscan_lidar_protobuf_obj in enumerate(self.images):
            
            processed_msg = from_proto(sickscan_lidar_protobuf_obj)
            # systemtimestamp = self.system_timestamps[index]
            # self.total_messages += 1
            
            # print(f"Amiga Timestamp (converted): {systemtimestamp}")
            # self.msg_timestamp = msg.header.timestamp_sec + msg.header.timestamp_nsec / 1e9 
            # self.msg_data_size = msg.data.size
            # self.msg_data_buffer = msg.data.buffer
            # self.msg_height = msg.height
            # self.msg_width = msg.width
            # npoints =  self.msg_height * self.msg_width
            # self.msg_num_fields = msg.fields.size

            result = self.py_sick_scan_cartesian_point_cloud_msg_to_xy(
                            processed_msg
            )
            if result is not None:
                x_points, y_points, z_points, intensity_points = result
                self.total_points_x.extend(x_points)
                self.total_points_y.extend(y_points)
                self.total_points_z.extend(z_points)
                self.total_points_intensity.extend(intensity_points)
                
                if index == 1:
                    shared_logger.info(f"The type of x is {(x_points.shape)}")

                self.total_timestamps.extend(
                    [
                        self.system_timestamps[index]
                    ] * len(x_points)
                )
                self.total_lidar_timestamps.extend(
                    [
                        self.msg_timestamp[index]
                    ] * len(x_points)
                )
        
        end_time = time.time()
        # Print elapsed time
        total_points = len(self.total_points_x)
        shared_logger.info(f"SickScan total messages: {self.total_messages} ; Total xyz points: {total_points} ; Points per message: {total_points / self.total_messages}")
        shared_logger.info(f"Processing time sickscan_lidar_protobuf): {end_time - start_time:.4f} seconds")
        check_memory_usage('extract 4')
        
        
        start_time = time.time() 
        self.filtered_data = False
        csv_path_name_ext = fpath / self.get_output_name(index = None, ext = "csv", details = "testing")
        header = ["Amiga_timestamp", "Lidar_timestamp", "X", "Y", "Z", "Intensity"]
        with open(
            csv_path_name_ext, "w", encoding="utf-8", newline=""
        ) as csv_file:
            writer = csv.writer(csv_file)
            writer.writerow(header)  # Write the header row

            if self.filtered_data:
                for system_time, lidar_time, x, y in zip(
                    self.total_timestamps,
                    self.total_lidar_timestamps,
                    self.total_points_x,
                    self.total_points_y,
                ):
                    writer.writerow([system_time, lidar_time, x, y])
            else:
                for system_time, lidar_time, x, y, z, intensity in zip(
                    self.total_timestamps,
                    self.total_lidar_timestamps,
                    self.total_points_x,
                    self.total_points_y,
                    self.total_points_z,
                    self.total_points_intensity,
                ):
                    writer.writerow(
                        [system_time, lidar_time, x, y, z, intensity]
                    )
            
        end_time = time.time()
        # shared_logger.info(f"Saving LIDAR data: {image_path_name_ext}  {utc_datetime}")
        shared_logger.info(f"Write time CSV sickscan_lidar_protobuf): {end_time - start_time:.4f} seconds")
        check_memory_usage('extract 5')    
            
        
    def py_sick_scan_cartesian_point_cloud_msg_to_xy_original(
        self, pointcloud_msg, start_time=None
    ):
        """
        This method converts the pointcloud_msg to x,y coordinates.
        
        Taken from the Resonate resonatesystems-rs24005-appn-instrument-interfaces git (Bitbucket) 
        project process_lidar_binaryfiles.py file.
        """

        num_fields = pointcloud_msg.fields.size
        msg_fields_buffer = pointcloud_msg.fields.buffer

        # print(f"Num of fields {num_fields}")
        # print(f"msg fields buffer is {msg_fields_buffer}")

        # Initialize offsets to None or some default value
        field_offset_x = -1
        field_offset_y = -1
        field_offset_z = -1
        field_offset_intensity = -1

        for n in range(num_fields):
            field_name = ctypesCharArrayToString(msg_fields_buffer[n].name)
            field_offset = msg_fields_buffer[n].offset
            if field_name == "x":
                field_offset_x = msg_fields_buffer[n].offset
            elif field_name == "y":
                field_offset_y = msg_fields_buffer[n].offset
            elif field_name == "z":
                field_offset_z = msg_fields_buffer[n].offset
            elif field_name == "intensity":
                field_offset_intensity = msg_fields_buffer[n].offset

        if (
            field_offset_x is None
            or field_offset_y is None
            or (not self.filtered_data and field_offset_z is None)
        ):
            raise ValueError("Offsets not assigned correctly.")



        cloud_data_buffer_len = pointcloud_msg.row_step * pointcloud_msg.height

        assert (
            pointcloud_msg.data.size == cloud_data_buffer_len
            and field_offset_x >= 0
            and field_offset_y >= 0
            and field_offset_intensity >= 0
            and (self.filtered_data or field_offset_z >= 0)
        )
        cloud_data_buffer = bytearray(cloud_data_buffer_len)

        for n in range(cloud_data_buffer_len):
            cloud_data_buffer[n] = pointcloud_msg.data.buffer[n]

        points_x = np.zeros(
            pointcloud_msg.width * pointcloud_msg.height, dtype=np.float32
        )
        points_y = np.zeros(
            pointcloud_msg.width * pointcloud_msg.height, dtype=np.float32
        )
        points_intensity = np.zeros(
            pointcloud_msg.width * pointcloud_msg.height, dtype=np.float32
        )
        points_z = None
        if not self.filtered_data:
            points_z = np.zeros(
                pointcloud_msg.width * pointcloud_msg.height, dtype=np.float32
            )
        point_idx = 0

        for row_idx in range(pointcloud_msg.height):
            for col_idx in range(pointcloud_msg.width):
                pointcloud_offset = (
                    row_idx * pointcloud_msg.row_step
                    + col_idx * pointcloud_msg.point_step
                )

                try:
                    points_x[point_idx] = np.frombuffer(
                        cloud_data_buffer,
                        dtype=np.float32,
                        count=1,
                        offset=pointcloud_offset + field_offset_x,
                    )[0]
                    points_y[point_idx] = np.frombuffer(
                        cloud_data_buffer,
                        dtype=np.float32,
                        count=1,
                        offset=pointcloud_offset + field_offset_y,
                    )[0]
                    if points_z is not None:
                        points_z[point_idx] = np.frombuffer(
                            cloud_data_buffer,
                            dtype=np.float32,
                            count=1,
                            offset=pointcloud_offset + field_offset_z,
                        )[0]
                    points_intensity[point_idx] = np.frombuffer(
                        cloud_data_buffer,
                        dtype=np.float32,
                        count=1,
                        offset=pointcloud_offset + field_offset_intensity,
                    )[0]
                except Exception as e:
                    print(f"Error reading buffer at point index {point_idx}: {e}")

                point_idx += 1

        if points_z is not None:
            return points_x, points_y, points_z, points_intensity
        else:
            return points_x, points_y, points_intensity
        
        
    



from pathlib import Path
from phenomate_core.preprocessing.lidar2d import lidar_pb2
import sys
import os

import ctypes 
import datetime

import phenomate_core.preprocessing.lidar2d.sick_scan_api as ss

def from_proto(protocol_message):

    header = ss.SickScanHeader(
        seq=protocol_message.header.seq,
        timestamp_sec=protocol_message.header.timestamp_sec,
        timestamp_nsec=protocol_message.header.timestamp_nsec,
        frame_id=protocol_message.header.frame_id,
    )

    num_fields = protocol_message.fields.size
    array = ss.SickScanPointFieldMsg * num_fields
    elements = array()

    for n in range(num_fields):
        field = ss.SickScanPointFieldMsg(
            name=protocol_message.fields.buffer[n].name,
            offset=protocol_message.fields.buffer[n].offset,
            datatype=protocol_message.fields.buffer[n].datatype,
            count=protocol_message.fields.buffer[n].count,
        )
        elements[n] = field

    fields = ss.SickScanPointFieldArray(
        capacity=protocol_message.fields.capacity,
        size=protocol_message.fields.size,
        buffer=elements,
    )

    byte_data = bytes(protocol_message.data.buffer)
    size = len(byte_data)
    buffer_copy = bytearray(byte_data)
    buffer_type = ctypes.c_uint8 * size
    buffer_instance = buffer_type.from_buffer(buffer_copy)
    # print(f"Buffer size is {size}")

    data = ss.SickScanUint8Array(
        capacity=protocol_message.data.capacity,
        size=protocol_message.data.size,
        buffer=buffer_instance,
    )

    message_contents = ss.SickScanPointCloudMsg(
        header=header,
        height=protocol_message.height,
        width=protocol_message.width,
        fields=fields,
        is_bigendian=protocol_message.is_bigendian,
        point_step=protocol_message.point_step,
        row_step=protocol_message.row_step,
        data=data,
        is_dense=protocol_message.is_dense,
        num_echos=protocol_message.num_echos,
        segment_idx=protocol_message.segment_idx,
    )

    return message_contents




from __future__ import annotations

import logging
import traceback
import struct
import time
from datetime import UTC, datetime
from pathlib import Path
from typing import Any
import shutil

from google.protobuf.message import DecodeError

import cv2
import tifffile
from PIL import Image, PngImagePlugin, __version__

from phenomate_core.get_version import get_version
from phenomate_core.preprocessing.base import BasePreprocessor
from phenomate_core.preprocessing.jai import jai_pb2

# shared_logger = logging.getLogger("celery")
# from celery.utils.log import get_task_logger
# shared_logger = get_task_logger(__name__)

from phenomate_core.get_version import get_task_logger
shared_logger = get_task_logger(__name__)

class JaiPreprocessor(BasePreprocessor[jai_pb2.JAIImage]):
    r"""Average Timing  and compression results (per image for 17 images) extracted from
    protobuffer file and saved with equivalent metadata.
    +-----------+----------------------------------+------------------+-----------------+-------------+------------------------+
    | Library   | Compression Method               | Write Time (s)   | Read Time (s)   | Size (MB)   | Notes                  |
    +===========+==================================+==================+=================+=============+========================+
    | tifffile  | none (not bigtiff)               | 0.0248           | 0.0329          | 35.53       |                        |
    +-----------+----------------------------------+------------------+-----------------+-------------+------------------------+
    | tifffile  | zstd (not bigtiff)               | 0.0689           | 0.0616          | 32.65       | Unreadable on Windows  |
    +-----------+----------------------------------+------------------+-----------------+-------------+------------------------+
    | tifffile  | jpeg2000 (not bigtiff)           | 1.2594           | 1.0774          | 17.71       | Unreadable on Windows  |
    +-----------+----------------------------------+------------------+-----------------+-------------+------------------------+
    | tifffile  | webp (not bigtiff)               | 4.6053           | 0.1284          | 19.65       | Unreadable on Windows  |
    +-----------+----------------------------------+------------------+-----------------+-------------+------------------------+
    | tifffile  | lzma (not bigtiff)               | 3.4209           | 0.4254          | 27.18       | Unreadable on Windows  |
    +-----------+----------------------------------+------------------+-----------------+-------------+------------------------+
    | PIL PNG   | png_pil (compression level 9)    | 2.8894           | 0.4572          | 21.76       |                        |
    +-----------+----------------------------------+------------------+-----------------+-------------+------------------------+
    | PIL PNG   | png_pil (compression level 0)    | 0.6047           | 0.2834          | 35.59       |                        |
    +-----------+----------------------------------+------------------+-----------------+-------------+------------------------+
    | PIL TIFF  | raw                              | 0.0455           | 0.0341          | 35.59       |                        |
    +-----------+----------------------------------+------------------+-----------------+-------------+------------------------+
    | PIL TIFF  | tiff_lzw                         | 0.5706           | 0.1503          | 40.94       |  Increases the size    |
    +-----------+----------------------------------+------------------+-----------------+-------------+------------------------+
    | PIL TIFF  | tiff_deflate                     | 0.9058           | 0.0992          | 32.47       |                        |
    +-----------+----------------------------------+------------------+-----------------+-------------+------------------------+

    As the tifffile library has the fastest read and write speed for uncompressed
    data I am going to use it as the default format. This gives the benefit that
    bigTIFF file support will also be available, which will be useful in the case
    that large stiched images result in >4GB data sizes.


    ToDo: Confirmation of copyright note is needed.
    ToDo: Further metadata should be added (Username / Institution ROR


    Some further Tiff tags of interest:
    # See site-packages\tifffile\tifffile.py line 16110
    # (4869, 'AndorTemperature'),
    # (4876, 'AndorExposureTime'),
    # (4878, 'AndorKineticCycleTime'),
    # (4879, 'AndorAccumulations'),
    # (4881, 'AndorAcquisitionCycleTime'),
    # (4882, 'AndorReadoutTime'),
    # (4884, 'AndorPhotonCounting'),
    # (4885, 'AndorEmDacLevel'),
    # (4890, 'AndorFrames'),
    # (4896, 'AndorHorizontalFlip'),
    # (4897, 'AndorVerticalFlip'),
    # (4898, 'AndorClockwise'),
    # (4899, 'AndorCounterClockwise'),
    # (4904, 'AndorVerticalClockVoltage'),
    # (4905, 'AndorVerticalShiftSpeed'),
    # (4907, 'AndorPreAmpSetting'),
    # (4908, 'AndorCameraSerial'),
    # (4911, 'AndorActualTemperature'),
    # (4912, 'AndorBaselineClamp'),
    # (4913, 'AndorPrescans'),
    # (4914, 'AndorModel'),
    # (4915, 'AndorChipSizeX'),
    # (4916, 'AndorChipSizeY'),
    # (4944, 'AndorBaselineOffset'),
    # (4966, 'AndorSoftwareVersion'),

    # | Tag Name               | Tag ID (Hex) | Description                                 |
    # |------------------------|--------------|---------------------------------------------|
    # | `GPSLatitudeRef`       | `0x0001`     | North or South latitude indicator (`N`/`S`) |
    # | `GPSLatitude`          | `0x0002`     | Latitude in degrees, minutes, seconds       |
    # | `GPSLongitudeRef`      | `0x0003`     | East or West longitude indicator (`E`/`W`)  |
    # | `GPSLongitude`         | `0x0004`     | Longitude in degrees, minutes, seconds      |
    # | `GPSAltitudeRef`       | `0x0005`     | Altitude reference (0 = above sea level)    |
    # | `GPSAltitude`          | `0x0006`     | Altitude in meters                          |
    # | `GPSTimeStamp`         | `0x0007`     | Time of GPS fix                             |
    # | `GPSDateStamp`         | `0x001D`     | Date of GPS fix                             |

    # {
      # "GPSLatitudeRef": "N",
      # "GPSLatitude": [34, 3, 30.12],
      # "GPSLongitudeRef": "E",
      # "GPSLongitude": [118, 14, 55.32]
    # }
    """

    def extract(self, **kwargs: Any) -> None:
        
        dir_part = self.path.parent  # this is another Path type
        file_part = self.path.name  # this is a str
        
        shared_logger.info(f"JaiPreprocessor.extract(): file_part: {file_part}")
        

        # Read the path that was written to the self.path+'.origin' file        
        origin_path = self.open_origin_file()  
        # Select the matching files from the path by the timestamp of the self.path file
        path_objects = self.matched_file_list(origin_path, file_part)
        
        if path_objects != None:
            self.extra_files = path_objects
            shared_logger.info(f"JaiPreprocessor.extract(): number of related files:  {len(self.extra_files)}")
            shared_logger.info(f"JaiPreprocessor.extract(): related files: {self.extra_files}")
        
        with self.path.open("rb") as file:
            while True:
                try:
                    # Read the length of the next serialized message
                    serialized_timestamp = file.read(8)
                    if not serialized_timestamp:
                        break
                    system_timestamp = struct.unpack("d", serialized_timestamp)[0]

                    length_bytes = file.read(4)
                    if not length_bytes:
                        break
                    length = int.from_bytes(length_bytes, byteorder="little")

                    # Read the serialized message
                    serialized_image = file.read(length)

                    # Parse the protobuf message
                    image_protobuf_obj = jai_pb2.JAIImage()
                    image_protobuf_obj.ParseFromString(serialized_image)

                    # Update to extracted image list
                    self.images.append(image_protobuf_obj)
                    self.system_timestamps.append(system_timestamp)

                    shared_logger.debug(
                        f"JaiPreprocessor.extract(): Converted timestamp: image.timestamp: {image_protobuf_obj.timestamp} framerate: {image_protobuf_obj.frame_rate}"
                    )
                except DecodeError as e:
                    shared_logger.exception(f"JaiPreprocessor.extract(): Protobuffer  Decode error for file: {file_part}: {e}")
                    break
                except Exception as e:
                    shared_logger.exception(f"JaiPreprocessor.extract(): Unexpected error while reading {self.path}: {e}")
                    raise
        shared_logger.info(f"JaiPreprocessor.extract() Number of images extraced:  {len(self.images)}")
        
  
    def copy_extra_files(self, fpath: Path) -> None:
        """
        Extra files that are associated with the .bin proto buffer data can be copied to the destination directory.
        
        :fpath Path: is the directory in which to save the files
        
        This method impicitly uses the extra_files list that should be populated in the extract() method using
        open_origin_file() and matched_file_list() (see: ImuPreprocessor.extract())
        """
        for file_path in self.extra_files:
            try:
                
                # For the JAI there should be 2 extra files, both json
                if file_path.suffix.lower() == ".json":
                    file_path = Path(file_path)
                    
                    details = ''
                    if "device_params".lower() in file_path.name.lower():
                        details = 'device_params'
                    if "stream_params".lower() in file_path.name.lower():
                        details = 'stream_params'
                    
                    output_file = file_path.name  # str
                    file_path_name_ext = fpath / self.get_output_name(
                            index=None, ext="json", details=details
                    ) 
                    shutil.copy(file_path, file_path_name_ext)
                    shared_logger.info(f"BasePreprocessor.copy_extra_files(): JAI data transfer: Copied file: {file_path_name_ext}") 

            except FileNotFoundError as e:
                shared_logger.error(f"BasePreprocessor: data transfer: File not found: {file_path} — {e}")
            except PermissionError as e:
                shared_logger.error(f"BasePreprocessor: data transfer: Permission denied: {file_path} — {e}")
            except OSError as e:
                shared_logger.error(f"BasePreprocessor: data transfer: OS error while accessing {file_path}: {e}")
            except Exception as e:
                shared_logger.exception(f"BasePreprocessor: data transfer: Unexpected error while reading {file_path}: {e}")
                raise

    def matched_file_list(self, origin_path: Path, file_part : str) -> list[Path]:
        """
        Return a list of files from the source directory that match the timstamp of the :path: file.
           
        :param origin_path: The path to the source directoy
        :type origin_path: Path
        :param file_part: The name of the selected data file, with a timestamp in the name
        :type file_part: str

        """
        # Set of all files in the directory
        files_in_dir = self.list_files_in_directory(origin_path.parent)
        shared_logger.debug(f"BasePreprocessor: files_in_dir:  {files_in_dir}")
        
        matched = []
        json_files = [f for f in files_in_dir if f.lower().endswith(".json")]
        shared_logger.debug(f"BasePreprocessor: json_files:  {json_files}")
        
        # Separate into two lists
        stream_params_files = [f for f in json_files if "stream_params" in f]
        device_params_files = [f for f in json_files if "device_params" in f]

        res_match = self.return_closest_in_time(stream_params_files, file_part)  
        if res_match != None:            
            matched.append(res_match)
        res_match = self.return_closest_in_time(device_params_files, file_part)  
        if res_match != None:            
            matched.append(res_match)

        
        if len(matched) > 0:
            matched_with_dir = [origin_path.parent / f for f in matched]
            path_objects = [Path(p) for p in matched_with_dir]
            
            shared_logger.info(f"BasePreprocessor:matched_file_list() path_objects files: {path_objects}")
            return path_objects 
        else:
            return None
            
        
    # Set bigtiff=True, for 64 bit TIFF  tags
    def save(
        self,
        path: Path | str,
        width: int | None = None,
        height: int | None = None,
        **kwargs: Any,
    ) -> None:
        """Save the data using the tifffile package.
        N.B. Only the Compression='none' output files are read natively by Windows
        """
        fpath = Path(path)
        fpath.mkdir(parents=True, exist_ok=True)
        shared_logger.info(f"JaiPreprocessor.save() output path (fpath): {fpath} ")
        
        start_time = time.time()
        self.copy_extra_files(fpath)
        # End timer
        end_time = time.time()
        # Print elapsed time
        shared_logger.info(f"JaiPreprocessor.save() Copy file time (JAI data): {end_time - start_time:.4f} seconds")
        
        
        
        current_year = str(datetime.now(UTC).year)
        phenomate_version = get_version()
        user = '''"creator": {
"@type": "Organization",
"name": "Australian Plant Phenomics Network",
"identifier": "https://ror.org/02zj7b759"
}'''  # 315 Creator of the image
        start_time = time.time()
        for index, image in enumerate(self.images):
            # Determine width and height
            iwidth = width if width is not None else image.width
            iheight = height if height is not None else image.height
            bayer_image = self.bytes_to_numpy(image.image_data).reshape((iheight, iwidth))

            # Conversion to use after discussion in #https://github.com/aus-plant-phenomics-network/phenomate-core/issues/2
            # rgb_image = cv2.cvtColor(bayer_image, cv2.COLOR_BayerRGGB2BGR)  # Use this if saving with cv2.imwrite
            rgb_image = cv2.cvtColor(bayer_image, cv2.COLOR_BayerRGGB2RGB)

            # utc_now = datetime.now(UTC)
            tiff_date = datetime.now(UTC).strftime("%Y:%m:%d %H:%M:%S")
            utc_datetime = datetime.fromtimestamp(image.timestamp / 1000000, tz=UTC)
            # shared_logger.info(f"Converted timestamp (no compression): image.timestamp: {image.timestamp}  {utc_datetime}")

            tag_269 = f'"title":"Phenomate JAI output",  "software": "phenomate-core {phenomate_version}", '
            tag_270 = '"A plant phenotype experiment image. Source image is JAI camera protobuffer object raw Bayer image. Output converted using OpenCV.cvtColor() and saved using the tifffile library"'
            tag_274 = tifffile.ORIENTATION.TOPLEFT  # ORIENTATION should be an integer value
            # tag_305 = # tifffile adds its own name here.
            tag_306 = f'{tiff_date}'
            tag_315 = f'{user}'
            tag_33432 = f'"Copyright {current_year} Australian Plant Phenomics Network. All rights reserved"'
            tag_65000 = '{ "timestamp_description": "system_timestamp"" : "The system timestamp that the image was added to the protocol buffer", "jai_collection_timestamp": "The JAI camera counter value when the image was taken" }'
            tag_65001 = f'{{ "system_timestamp": "{self.system_timestamps[index]}" }}'
            tag_65002 = f'{{ "jai_collection_timestamp": "{image.timestamp}" }} '

            extratags = [
                (269, "s", len(tag_269) + 1, tag_269, True),  # 269 DocumentName
                # (270, 's', len(tag_270) + 1, tag_270, True),      # Use the description parameter in the tifffile.imwrite() method
                (274, "I", 1, tag_274, True),  # 274 Image orientation
                # (305, 's', len(tag_305) + 1, tag_305, True),      # 305 software version - tifffile adds its own name here.
                (306, "s", len(tag_306) + 1, tag_306, True),  # 306 Creation time
                (315, "s", len(tag_315) + 1, tag_315, True),  # 315 Creator of the image
                (33432, "s", len(tag_33432) + 1, tag_33432, True),  # 33432 Copyright information
                (65000, "s", len(tag_65000) + 1, tag_65000, True),
                # (65001, 'Q', 1, image.timestamp, True),           # For 64 bit tags are enabled by bigtiff=True
                (65001, "s", len(tag_65001) + 1, tag_65001, True),
                (65002, "s", len(tag_65002) + 1, tag_65002, True),
            ]

            compression_l = "none"  # lossless: lzma  zstd   compressionargs={'lossless': True} not available: bzip2 lz4 ; slow: jpeg2000, webp

            image_path_name_ext = fpath / self.get_output_name(
                index=image.timestamp, ext="tiff", details=None
            )
          
            try:
                tifffile.imwrite(
                    f"{image_path_name_ext}",
                    rgb_image,
                    bigtiff=False,
                    planarconfig="contig",  # This is the default interleaved rgb format.
                    compression=compression_l,
                    # compression='jpeg | jpeg2000' , compressionargs={'level': 100},   # JPEG quality level (0 to 100) 0 is lower quality
                    # compressionargs={'lossless': True},  # webp quality level
                    description=tag_270,
                    extratags=extratags,
                    photometric="rgb",
                )
            except IOError as e:
                shared_logger.error(f"JaiPreprocessor.save() I/O error occurred: {e}")
                raise
            except FileNotFoundError:
                shared_logger.error("JaiPreprocessor.save() File not found.")            
                raise
            except PermissionError:
                shared_logger.error("You do not have permission to write to this file.")
                raise
            except Exception as e:
                shared_logger.error(f"Unexpected error: {e}")
                shared_logger.error(f"JaiPreprocessor.save() {traceback.format_exc()}")
                raise

        end_time = time.time()
        # Print elapsed time
        shared_logger.info(
            f"JaiPreprocessor.save() Write time for {index+1} files: (tifffile compression: {compression_l}, not bigtiff): {end_time - start_time:.4f} seconds "
        )

    # PNG data conversion code using PIL save_png_with_metadata_with_PIL()
    def save_png_with_metadata_with_PIL(
        self,
        path: Path | str,
        width: int | None = None,
        height: int | None = None,
        **kwargs: Any,
    ) -> None:
        """PNG format is lossless and the high compression ratio makes it a good archival
        format, however it is relatively slow, even without compression.
        """
        fpath = Path(path)
        fpath.mkdir(parents=True, exist_ok=True)

        png_compression = "0"
        png_lib = "pil"
        current_year = str(datetime.now(UTC).year)
        phenomate_version = get_version()
        start_time = time.time()
        for index, image in enumerate(self.images):
            # Determine width and height
            iwidth = width if width is not None else image.width
            iheight = height if height is not None else image.height
            bayer_image = self.bytes_to_numpy(image.image_data).reshape((iheight, iwidth))

            # Conversion to use after discussion in #https://github.com/aus-plant-phenomics-network/phenomate-core/issues/2
            # rgb_image = cv2.cvtColor(bayer_image, cv2.COLOR_BayerRGGB2BGR)  # Use this if saving with cv2.imwrite
            rgb_image = cv2.cvtColor(bayer_image, cv2.COLOR_BayerRGGB2RGB)

            utc_now = datetime.now(UTC)

            tag_270 = "A plant phenotype experiment image. Image taken by JAI camera protobuffer as a raw Bayer image and converted to standardised RGB using OpenCV cvtColor()"
            tag_274 = "ORIENTATION.TOPLEFT"  # ORIENTATION should be an integer value
            tag_305 = f"phenomate-core version: {phenomate_version} using Python library PNG writer: {png_lib}, version {__version__}"
            tag_306 = f"{utc_now}"
            user = '''"creator": {
"@type": "Organization",
"name": "Australian Plant Phenomics Network",
"identifier": "https://ror.org/02zj7b759"
}'''   # 315 Creator of the image
            tag_315 = f"{user}"
            tag_33432 = (
                f"Copyright {current_year} Australian Plant Phenomics Network. All rights reserved."
            )
            tag_65500 = 'timestamp_description: "System_timestamp: timestamp from when the image was added to the protocol buffer", "JAI_collection_timestamp: JAI counter value when the image was taken" } '
            tag_65501 = f"{self.system_timestamps[index]}"
            tag_65502 = f"{image.timestamp}"

            # Create a PngInfo object to hold metadata
            metadata = PngImagePlugin.PngInfo()
            metadata.add_text("Timestamp_Info", tag_65500)
            metadata.add_text("System_timestamp", tag_65501)
            metadata.add_text("JAI_collection_timestamp", tag_65502)
            metadata.add_text("Description", tag_270)
            metadata.add_text("Orientation", tag_274)
            metadata.add_text("Software", tag_305)
            metadata.add_text("Current_Time", tag_306)
            metadata.add_text("Author", tag_315)
            metadata.add_text("Copyright", tag_33432)

            out_image = Image.fromarray(rgb_image)
            png_compression = 0 if png_compression == "none" else int(png_compression)
            image_path_name_ext = fpath / self.get_output_name(
                image.timestamp, "png", f"compress{png_compression}_{png_lib}_"
            )
            out_image.save(
                image_path_name_ext, format="PNG", pnginfo=metadata, compress_level=png_compression
            )

        # End timer
        end_time = time.time()
        # Print elapsed time
        shared_logger.info(
            f"Write time ({png_lib} {png_compression} compression tiff): {end_time - start_time:.4f} seconds"
        )

    # The 32 bit TIFF PIL writer code save_tiff_with_PIL()
    def save_tiff_with_PIL(
        self,
        path: Path | str,
        width: int | None = None,
        height: int | None = None,
        **kwargs: Any,
    ) -> None:
        """Tiff writer with added Tag metadata and custom Tag values 65001, 65002 for image collection timestamps

        # Timing for 17 images :
        # Write time (tiff_lzw pil tiff)    : 9.6995 seconds;  reading (lzw_pil.tiff images)    : 2.5557 seconds; 696MB
        # Write time (raw pil tiff)         : 0.7733 seconds;  reading (raw_pil.tiff images)    : 0.5805 seconds; 605MB
        # Write time (tiff_deflate pil tiff): 15.3987 seconds; reading (deflate_pil.tiff images): 1.6871 seconds; 552MB
        """
        fpath = Path(path)
        fpath.mkdir(parents=True, exist_ok=True)
        tiff_compression = "tiff_deflate"  # "tiff_deflate" "raw" "jpeg"
        image_lib = "pil"
        current_year = str(datetime.now(UTC).year)
        phenomate_version = get_version()
        start_time = time.time()
        for index, image in enumerate(self.images):
            # Determine width and height
            iwidth = width if width is not None else image.width
            iheight = height if height is not None else image.height
            bayer_image = self.bytes_to_numpy(image.image_data).reshape((iheight, iwidth))

            # Conversion to use after discussion in #https://github.com/aus-plant-phenomics-network/phenomate-core/issues/2
            # rgb_image = cv2.cvtColor(bayer_image, cv2.COLOR_BayerRGGB2BGR)  # Use this if saving with cv2.imwrite
            rgb_image = cv2.cvtColor(bayer_image, cv2.COLOR_BayerRGGB2RGB)

            # utc_now = datetime.now(timezone.utc)
            tiff_date = datetime.now(UTC).strftime(
                "%Y:%m:%d %H:%M:%S"
            )  # This is a required format for the tiff date/time tag

            tag_269 = '"Phenomate JAI output"'
            tag_270 = f'"A plant phenotype experiment image. Source image is JAI camera protobuffer object raw Bayer image. Output converted using OpenCV.cvtColor() and saved using the PIL library with compression: {tiff_compression}"'
            tag_274 = tifffile.ORIENTATION.TOPLEFT  # ORIENTATION should be an integer value
            tag_305 = f"phenomate-core version: {phenomate_version} written using Python library PIL version {__version__}"
            tag_306 = f"{tiff_date}"
            user = '''"creator": {
"@type": "Organization",
"name": "Australian Plant Phenomics Network",
"identifier": "https://ror.org/02zj7b759" 
}'''   # 315 Creator of the image  # 315 Creator of the image
            tag_315 = f"{user}"
            tag_33432 = (
                f"Copyright {current_year} Australian Plant Phenomics Network. All rights reserved"
            )
            tag_65000 = '{ "timestamp_description" : "system_timestamp is the time that the image was added to the protocol buffer; jai_collection_timestamp is the JAI camera counter value when the image was taken" }'
            tag_65001 = f"{self.system_timestamps[index]}"
            tag_65002 = f"{image.timestamp}"

            metadata = {
                269: tag_269,
                270: tag_270,
                274: tag_274,
                305: tag_305,
                306: tag_306,
                315: tag_315,
                33432: tag_33432,
                65000: tag_65000,
                65001: tag_65001,
                65002: tag_65002,
            }

            # Convert the reshaped image data to a PIL Image object
            out_image = Image.fromarray(rgb_image).convert("RGB")
            image_path_name_ext = fpath / self.get_output_name(
                image.timestamp, "tiff", f"{tiff_compression}_{image_lib}"
            )
            out_image.save(
                image_path_name_ext,
                format="TIFF",
                tiffinfo=metadata,
                compression=None if tiff_compression == "none" else tiff_compression,
            )  # tiff_adobe_deflate tiff_jpeg

        # End timer
        end_time = time.time()
        # Print elapsed time
        shared_logger.info(
            f"Write time ({tiff_compression} {image_lib} tiff): {end_time - start_time:.4f} seconds"
        )
